(*
 * The MIT License (MIT)
 * 
 * Copyright (c) 2015 Cole Schlesinger
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *)

open Compiler
open Syntax
open Test_compiler


(* Do compilation, then measure the size of the original vs. compiled policies. *)
let run_benchmark
  ((size, (name, p, tables)) : (int * (string * pol * ((TableName.t * pol) list))))
  : unit =

  (* Prepare substitutions. *)
  let b = mkTableBinding tables in
  let sub_p = sub_pol b p in

  (* Multicast consolidation. *)
  let mc_smap =
    List.fold_left
      (fun map (x, px) ->
        let metadata =
          MetadataField.gen_many (estimate_mc_metadata px) in
        TableBinding.add x metadata map)
      TableBinding.empty
      tables in
  let mc_s =
    (fun (x,t) ->
      try
        TableBinding.get x mc_smap
      with
        Not_found ->
          let msg = Printf.sprintf "mc_s: not found: (%s, %s)"
            (TableName.print x) (print_pol_t t) in
          raise (Failure msg)) in
  let mc_metadata =
    MetadataField.gen_many (estimate_mc_metadata p) in
  let m1, n1, theta1 = pipeline mc_s mc_metadata p mod_stage_id in
  let n1_pol = egress_stage_to_pol n1 in
  let n1_tables = tables_of_pol n1_pol in

  (* Field extraction. *)
  let b_after_mc = theta1 b in
  let (fe_smap : FieldSet.t TableBinding.t) =
    TableBinding.mapi
      (fun k (res, _) ->
        let pol = match res with
          | Policy p -> p
          | ModStage m -> mod_stage_to_pol m
          | EgressStage n -> egress_stage_to_pol n
          in
        if TableSet.mem k n1_tables then
          MetadataField.gen_many (estimate_fe_metadata pol)
        else
          FieldSet.empty)
      b_after_mc in
  let fe_s = (fun (x, _) -> TableBinding.get x fe_smap) in
  let fe_metadata =
    MetadataField.gen_many (estimate_fe_metadata n1_pol) in
  let e1, n2, theta2 =
    ext "OutPort" fe_s fe_metadata n1_pol mod_stage_id in

  (* Combine consolidation and extraction stages. *)
  let s1, s2, s3, theta3 = combine_mc_and_fe m1 e1 in

  (* Build the final policy. *)
  let p_out = s1 <.> s2 <.> s3 <.> n2 in

  (* Do substitution. *)
  let theta = (fun b -> (theta1 b) |> theta2 |> theta3) in
  let sub_m_out = 
    sub_pol_forced
      (theta b)
      ((mod_stage_to_pol m1) <.> (egress_stage_to_pol n1)) in
  let sub_me_out = 
    sub_pol_forced
      (theta b)
      ((mod_stage_to_pol m1) <.> (mod_stage_to_pol e1) <.> n2) in
  let sub_p_out = sub_pol_forced (theta b) p_out in

  let sub_s3 = sub_pol_forced (theta b) s3 in
  let sub_n2 = sub_pol_forced (theta b) n2 in
  let s3_tabled, s3_costs = insert_tables sub_s3 in
  let n2_tabled, n2_costs = insert_tables sub_n2 in
  let s3_num_tables = partition s3_tabled (s3_costs TableBinding.empty) 
    |> size_of_partition in
  let n2_num_tables = partition n2_tabled (n2_costs TableBinding.empty)
    |> size_of_partition in

  (* Print sizes. *)
  let p_size = size_of_pol sub_p in
  let p_m_size = size_of_pol sub_m_out in
  let p_me_size = size_of_pol sub_me_out in
  let p_out_size = size_of_pol sub_p_out in

  (* Estimate 3 operators per "row" in table. *)
  let single_table_size = (size * 3) * (size * 3) + 1 in

  (* Estimate number of physical tables to fit the one big logical table. *)
  let single_table_num_tables =
    (blocks sub_p TableBinding.empty) |/ 16 in

  (* <p size> <m size> <me size> <out size> <single size>
     <single tables> <out tables> *)
  Printf.printf "(%s) %d: %d\t%d\t%d\t%d\t%d\t%d\t%d\n"
    name size p_size p_m_size p_me_size p_out_size single_table_size
    single_table_num_tables (1 + s3_num_tables + n2_num_tables)

let setup_benchmark test size () =
  match test with
    | "learning" -> run_benchmark (size, learning_switch size)
    | "firewall" -> run_benchmark (size, stateful_firewall size)
    | _ -> raise (Failure (Printf.sprintf "unrecognized benchmark: %s" test))

let command =
  Core.Std.Command.basic
    ~summary:"Run one compiler benchmark."
    ~readme:(fun () -> Printf.sprintf "%s\n%s\n\n%s\n\n%s\n" 
        "TEST: 'learning' or 'firewall'"
        "SIZE: (int) size of source policy"
        "Size is calculated as syntactic policy size after substituting a population-time update.  It is displayed in the following format:"
        "(name) <virtual pipeline> <multicast consolidated> <field extracted> <physical pipeline>")
    Core.Std.Command.Spec.(empty +> anon ("test" %: string) +> anon ("size" %: int))
    setup_benchmark

let () =
  Core.Std.Command.run ~version:"1.0" ~build_info:"RWO" command
