(*
 * The MIT License (MIT)
 * 
 * Copyright (c) 2015 Cole Schlesinger
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *)

open My_core
open Syntax
open Equivalence

(** TYPES OF COMPILATION STAGES **********************************************)

(* Type of both consolidation and modification stages. *)
type mod_sequence =
  | Mods of pol list
  | TMods of (TableName.t * pol_t * pol list)
(* Sum of sequences.  The empty list/sum is equivalent to drop. *)
type mod_stage = (NF.t * mod_sequence) list

(* The "identity" consolidation stage, equivalent to Id, eg. used for the
 * top-level invokation of pipeline.*)
let mod_stage_id = [(nf_of_pred Id, Mods [])]

let print_mod_sequence (c : mod_sequence) : string =
  match c with
  | Mods ps -> string_join "; " (List.map print_pol ps)
  | TMods (x, t, ps) ->
    Printf.sprintf "(%s : %s); %s"
      (TableName.print x)
      (print_pol_t t)
      (string_join "; " (List.map print_pol ps))

let print_mod_stage (c : mod_stage) : string =
  string_join " + "
    (List.map
      (fun (a,s) ->
        Printf.sprintf "(%s; %s)" (print_nf a) (print_mod_sequence s)) c)

let mod_sequence_to_pol (c : mod_sequence) : pol =
  match c with
  | Mods ps -> List.fold_right (fun md acc -> md <.> acc) ps (Pred Id)
  | TMods (x, t, ps) ->
    let mods = List.fold_right (fun md acc -> md <.> acc) ps (Pred Id) in
    Table (x,t) <.> mods

let mod_stage_to_pol (c : mod_stage) : pol =
  List.fold_left
    (fun acc (a,s) -> (Pred (pred_of_nf a) <.> mod_sequence_to_pol s) <+> acc)
    (Pred Drop)
    c

let mod_stage_dist (a : pred) (c : mod_stage) (ps : pol list) : mod_stage =
  List.map (fun (a', seq) ->
    match seq with
    | Mods ps' -> (nf_and a a', Mods (List.concat [ps; ps']))
    | TMods (x, t, ps') -> (nf_and a a', TMods (x, t, List.concat [ps; ps'])))
    c

let mod_stage_add_mod
  (f : Field.t)
  (v : int)
  ((a,s) : NF.t * mod_sequence) : (NF.t * mod_sequence) =
  match s with
  | Mods ps -> (a, Mods ((Modify (f,v))::ps))
  | TMods (x,t,ps) -> (a, TMods (x, t, ((Modify (f,v))::ps)))

let mod_stage_concat = List.concat

(* Type of egress stages.*)
type egress_atom =
  | ETable of TableName.t * pol_t
  | EPolicy of pol
type egress_guarded_atom =
  | EAtom of egress_atom
  | EGuard of NF.t * egress_atom
type egress_stage = egress_guarded_atom list

(* Egress stage equivalent to the identity policy. *)
let egress_id = [EAtom (EPolicy (Pred Id))]

(* Egress stage equivalent to the drop policy. *)
let egress_drop = []

(* Egress stage equivalent to p. *)
let egress_pol (p : pol) : egress_stage = [EAtom (EPolicy p)]

(* Egress stage table. *)
let egress_table ((x,t) : TableName.t * pol_t) : egress_stage =
  [EAtom (ETable (x,t))]

let print_egress_atom (ea : egress_atom) : string =
  match ea with
  | ETable (x,t) ->
    Printf.sprintf "(%s : %s)" (TableName.print x) (print_pol_t t)
  | EPolicy r -> print_pol r

let print_egress_guarded_atom (eg : egress_guarded_atom) : string =
  match eg with
  | EAtom at -> print_egress_atom at
  | EGuard (nf, at) ->
    Printf.sprintf "if %s then %s else Id"
      (print_nf nf)
      (print_egress_atom at)

let print_egress_stage (e : egress_stage) : string =
  if List.length e = 0 then
    "Id"
  else
    string_join "; " (List.map print_egress_guarded_atom e)

let egress_atom_to_pol (ea : egress_atom) : pol =
  match ea with
  | ETable (x,t) -> Table (x,t)
  | EPolicy r -> r

let egress_guarded_atom_to_pol (eg : egress_guarded_atom) : pol =
  match eg with
  | EAtom at -> egress_atom_to_pol at
  | EGuard (nf, at) -> 
    let nf_pred = pred_of_nf nf in
    mkIf nf_pred (egress_atom_to_pol at) (Pred Id)

let egress_stage_to_pol (n : egress_stage) : pol =
  List.fold_left
    (fun acc ea -> (egress_guarded_atom_to_pol ea) <.> acc)
    (Pred Id)
    n

let egress_stage_concat = List.concat


(** TABLE BINDINGS AND TRANSFORMERS ******************************************)

type replacement =
  | Policy of pol
  | ModStage of mod_stage
  | EgressStage of egress_stage

module TableBindingMap = Map.Make( struct
  type t = TableName.t
  let compare = TableName.compare
  end )

module TableBinding = struct
  include TableBindingMap

  let get (x : TableName.t) (map : 'a t) =
    try
      find x map
    with
      Not_found ->
        let msg =
          Printf.sprintf
            "(TableBinding.get) not found: (%s)"
            (TableName.print x) in
        raise (Failure msg)

  let put
    (x : TableName.t) 
    ((rep, size) : replacement * int) 
    (map : (replacement  * int) t) =
    let p = match rep with
      | Policy p -> p
      | ModStage m -> mod_stage_to_pol m
      | EgressStage e -> egress_stage_to_pol e
      in
    if not (TableSet.is_empty (tables_of_pol p)) then
      let msg = Printf.sprintf "(TableBinding.put) binding with table (%s)" (print_pol p) in
      raise (Undefined msg)
    else
      add x (rep, size) map

  let to_pol_list
    (b : (replacement  * int) t)
    : (TableName.t * pol) list =
    List.map
      (fun (k, (rep, _)) ->
        let p = match rep with
        | Policy p -> p
        | ModStage m -> mod_stage_to_pol m
        | EgressStage e -> egress_stage_to_pol e
        in
      (k, p))
      (bindings b)
end

(* Maximum policy size that can replace a table. *)
type size_estimate = int
type table_binding = (replacement * size_estimate) TableBinding.t
type theta = table_binding -> table_binding

(* Estimate number of rows in pol. *)
let rec height
  (pol : pol)
  (table_estimates : table_binding) : int =
  let rec estimate_pred pr =
    match pr with
    | Id
    | Drop
    | Match _ -> 1
    | Neg a -> estimate_pred a
    | And (a,b)
    | Or (a,b) -> (estimate_pred a) * (estimate_pred b)
    in
  match pol with
  | Pred a -> estimate_pred a
  | Modify _ -> 1
  | Table (x,_) -> snd (TableBinding.get x table_estimates)
  | IfThen (a, p, q) ->
    (height (Pred a <.> p) table_estimates) +
    (height (Pred (Neg a) <.> q) table_estimates)
  | Union (p,q)
  | Seq (p,q) ->
    (height p table_estimates) *
    (height q table_estimates)
  | Con (p, _, _, q) ->
    (height p table_estimates) +
    (height q table_estimates)


(* t2 \circ t1 *)
let compose (t2 : theta) (t1 : theta) : theta = fun b -> t2 (t1 b)

(* Replace tables in p according to b.  Raise Undefined if p contains a table
 * not mapped by b or if the substitution isn't type preserving. *)
let rec sub_pol (b : table_binding) (p : pol) : pol =
  match p with
  | Pred _
  | Modify _ -> p
  | Table (x, t) ->
    let res, _ =
      try
        TableBinding.get x b
      with
        Not_found -> raise (Undefined "(sub_pol) table not in binding")
      in
    begin match res with
    | Policy q ->
      let tq = infer_pol q in
      if not (tq <: t) then
        raise (Undefined "(sub_pol) type mismatch between table and binding")
      else if pol_contains_table q then
        raise (Undefined "(sub_pol) bound policy contains table")
      else
        q
    | _ ->
      let msg = Printf.sprintf "(sub_pol) non-policy bound to policy table" in
      raise (Undefined  msg)
    end
  | IfThen (a,p1,p2) -> mkIf a (sub_pol b p1) (sub_pol b p2)
  | Union (p1, p2) -> sub_pol b p1 <+> sub_pol b p2
  | Seq (p1, p2) -> sub_pol b p1 <.> sub_pol b p2
  | Con (p1, w1, w2, p2) ->
    let p1' = sub_pol b p1 in
    let p2' = sub_pol b p2 in
    let tp1' = infer_pol p1' in
    let tp2' = infer_pol p2' in
    if (tp1' <: (FieldSet.empty, w1)) && (tp2' <: (FieldSet.empty, w2)) then
      Con (p1', w1, w2, p2')
    else
      raise (Undefined "(sub_pol) type mismatch between table and binding")

(* Like sub_pol_forced, but converts non-policy bindings (like ModStage/EgressStage)
 * to policies. *)
let rec sub_pol_forced (b : table_binding) (p : pol) : pol =
  match p with
  | Pred _
  | Modify _ -> p
  | Table (x, t) ->
    let res, _ =
      try
        TableBinding.get x b
      with
        Not_found ->
          let msg = Printf.sprintf "(sub_pol_forced) table not in binding (%s)" (TableName.print x) in
          raise (Undefined msg)
      in
    let q = begin match res with
      | Policy q -> q
      | ModStage m -> mod_stage_to_pol m
      | EgressStage e -> egress_stage_to_pol e
      end in
    let tq = infer_pol q in
    if not (tq <: t) then
      let msg = Printf.sprintf "(sub_pol_forced) type mismatch between binding and table: (%s) not <: (%s)"
        (print_pol_t tq) (print_pol_t t) in
      raise (Undefined msg)
    else if pol_contains_table q then
      raise (Undefined "(sub_pol_forced) bound policy contains table")
    else
      q
  | IfThen (a,p1,p2) -> mkIf a (sub_pol_forced b p1) (sub_pol_forced b p2)
  | Union (p1, p2) -> sub_pol_forced b p1 <+> sub_pol_forced b p2
  | Seq (p1, p2) -> sub_pol_forced b p1 <.> sub_pol_forced b p2
  | Con (p1, w1, w2, p2) ->
    let p1' = sub_pol_forced b p1 in
    let p2' = sub_pol_forced b p2 in
    let tp1' = infer_pol p1' in
    let tp2' = infer_pol p2' in
    if (tp1' <: (FieldSet.empty, w1)) && (tp2' <: (FieldSet.empty, w2)) then
      Con (p1', w1, w2, p2')
    else
      raise (Undefined "(sub_pol_forced) type mismatch between table and binding")


let rec sub_mods
  (b : table_binding)
  (c : mod_stage) : mod_stage =
  match c with
  | [] -> []
  | (a, Mods ps)::tl -> (a, Mods ps)::(sub_mods b tl)
  | (a, TMods (x, t, ps))::tl ->
    let res, _ =
      try
        TableBinding.get x b
      with
        Not_found -> raise (Undefined "(sub_mods) table not in binding")
      in
    let c = match res with
      | ModStage c ->
        let c_pol = mod_stage_to_pol c in
        let tc = infer_pol c_pol in
        if not (tc <: t) then
          let msg = Printf.sprintf "(sub_mods) (%s : %s) -> (%s) : (%s)"
            (TableName.print x)
            (print_pol_t t)
            (print_pol c_pol)
            (print_pol_t tc) in
          raise (Undefined msg)
        else if pol_contains_table c_pol then
          raise (Undefined "(sub_mods) bound policy contains table")
        else
          c
      | _ -> raise (Undefined "(sub_mods) non-cons stage bound to cons stage table")
      in
    let c' = mod_stage_dist (pred_of_nf a) c ps in
    let rest = sub_mods b tl in
    List.concat [c'; rest]

(* Add a (and guards) to n.  (Definition 10.) *)
let qualify (a : pred) (n : egress_stage) : egress_stage =
  let qualify_guarded_atom (eg : egress_guarded_atom) : egress_guarded_atom =
    match eg with
    | EAtom at -> EGuard (nf_of_pred a, at)
    | EGuard (nf, at) -> EGuard (nf_and a nf, at)
    in
  if List.length n = 0 then
    [EGuard (nf_of_pred a, EPolicy (Pred Id))]
  else
    List.map qualify_guarded_atom n

let sub_egress (b : table_binding) (n : egress_stage) : egress_stage =
  let sub_atom (ea : egress_atom) : egress_stage =
    match ea with
    | ETable (x,t) ->
      begin try
        let n' = match TableBinding.get x b with
          | EgressStage n'', _ -> n''
          | _, _ -> raise (Undefined "(sub_egress) non-egress bound to egress table")
          in
        let p = egress_stage_to_pol n' in
        let tp = infer_pol p in
        if not (tp <: t) then
          raise (Undefined
            "(sub_egress) type mismatch between table and binding")
        else if pol_contains_table p then
          raise (Undefined
            "(sub_egress) bound policy contains table")
        else
          n'
      with
        Not_found -> raise (Undefined "(sub_egress) table not in binding")
      end
    | EPolicy r ->
      (* Sanity check: make sure r doesn't contain any tables. *)
      begin try
        let _ = sub_pol_forced TableBinding.empty r in
        [EAtom (EPolicy r)]
      with
        Undefined _ ->
        let msg =
          Printf.sprintf "(sub_egress) policy contains tables: (%s)"
            (print_egress_atom ea) in
        raise (Undefined msg)
      end
    in
  let sub_guarded_atom (eg : egress_guarded_atom) : egress_stage =
    match eg with
    | EAtom at -> sub_atom at
    | EGuard (nf, at) -> 
      let n' = sub_atom at in
      qualify (pred_of_nf nf) n'
    in
  List.flatten (List.map sub_guarded_atom n)


(* TODO: check table binding well-formedness. *)


(** HELPER FUNCTIONS *********************************************************)

(* Produce a predicate b such that f = v; a == b; f = v. *)
let rec specialize (f : Field.t) (v : int) (a : pred) : pred =
  match a with
  | Id
  | Drop -> a
  | Match (f', v') ->
    if f = f' && v = v' then
      (* Slight optimization: a;a == a. *)
      Id
    else if f = f' && v != v' then
      Drop
    else
      a
  | Neg b -> (<!>) (specialize f v b)
  | And (b1, b2) -> specialize f v b1 <^> specialize f v b2
  | Or (b1, b2) -> specialize f v b1 <||> specialize f v b2

let specialize_across_mods (b : pred) (mods : mod_sequence) : pred =
  let get_fvs modspol = match modspol with
    | Modify (f,v) -> (f,v)
    | _ ->
      raise
      (Undefined "(specialize_across_mods) Mods contains non-mod, non-seq policy")
    in
  let mods_list = match mods with
    | Mods pols -> List.map get_fvs pols
    | TMods _ ->
      raise
      (Undefined "(specialize_across_mods) cannot specialize across tables")
    in
  List.fold_left (fun acc (f,v) -> specialize f v acc) b mods_list



(** MULTICAST CONSOLIDATION **************************************************)

(* Estimate the number of metadata fields necessary for multicast
 * consolidation, not including tables. *)
let rec estimate_mc_metadata (p : pol) : int =
  match p with
  | Pred _
  | Modify _ 
  | Table _ -> 0
  | IfThen (_, p1, p2)
  | Seq (p1, p2)
  | Con (p1, _, _, p2) -> (estimate_mc_metadata p1) + (estimate_mc_metadata p2)
  | Union (p1, p2) -> (estimate_mc_metadata p1) + (estimate_mc_metadata p2) + 1

(* Split a fieldset into two subsets s1 and s2 such that |s1| >= n and |s2| >=
 * m.  Raise Failure "field split" if the set does not contain enough fields.
 *)
let split_fields
  (fs : FieldSet.t)
  (n : int)
  (m : int) : (FieldSet.t * FieldSet.t) =
  let fs_list = FieldSet.elements fs in
  let left, right = take n fs_list in
  if List.length right < m then
    raise (Failure "split_fields")
  else
    (mkFieldSet left, mkFieldSet right)

let split_metadata
  (fs : FieldSet.t)
  (estimate : pol -> int)
  (p : pol)
  (q : pol) : (FieldSet.t * FieldSet.t) =
  let metadata_for_p = estimate p in
  let metadata_for_q = estimate q in
  split_fields fs metadata_for_p metadata_for_q
  

(* Multicast consolidation.  (Definitions 7 and 8.) *)
let rec pipeline
  (s : (TableName.t * pol_t) -> FieldSet.t)
  (metadata : FieldSet.t)
  (pol : pol)
  (m : mod_stage) : (mod_stage * egress_stage * theta) =
  match pol, m with

  (* pipeline s p drop *)
  | _, [] -> ([], egress_drop, (fun x -> x))

  (* pipeline s id m *)
  | Pred Id, m -> (m, egress_id, (fun x -> x))

  (* pipeline s drop m *)
  | Pred Drop, _ -> ([], egress_drop, (fun x -> x))

  (* pipeline s (f = v) (h::t) *)
  | Pred b, m ->
    let m' = List.map (fun (a,p) -> (nf_and b a, p)) m in
    (m', egress_pol pol, (fun x -> x))

  (* pipeline s (f <- v) (m + a;q) *)
  | Modify (f,v), (a, (Mods q))::m' ->
    let a' = specialize f v (pred_of_nf a) in
    let m1, n1, theta1 = pipeline s metadata pol m' in
    if pred_equiv (Match (f,v) <^> a') Drop then
      (m1, n1, theta1)
    else
      ((nf_of_pred a',(Mods q))::m1, egress_pol pol, theta1)

  (* pipeline s (f <- v) (m + a; (x : t); q) *)
  | Modify _, (a, (TMods (x, t, q)))::m' ->
    let con_fragment = TMods (x, t, q) in
    let (con_fragment_pol : pol) =
      Pred (pred_of_nf a) <.> mod_sequence_to_pol con_fragment in
    let t1 = infer_pol con_fragment_pol in
    let x1 = TableName.gen () in
    let theta1 (b : table_binding) =
      let _, x_size = TableBinding.get x b in
      let m1, _, _ = pipeline s metadata pol (sub_mods b [(a, con_fragment)]) in
      TableBinding.put x1 (ModStage m1, x_size) b
      in
    let m2, _, theta2 = pipeline s metadata pol m' in
    ((nf_of_pred Id, TMods (x1, t1, []))::m2, egress_pol pol, compose theta2 theta1)

  (* pipeline s (x : (R,W)) m *)
  | Table (x, (r,w)), m ->
    let fs = s (x, (r,w)) in
    let m_pol = mod_stage_to_pol m in
    let t = infer_pol m_pol in
    let xm = TableName.gen () in
    let xmt = pol_t_union (r, fs) t in
    let xn = TableName.gen () in
    let xnt = (FieldSet.union r fs, w) in
    let theta' (b : table_binding) =
      let _, x_size = TableBinding.get x b in
      let m', n, _ = pipeline s fs (sub_pol_forced b pol) (sub_mods b m) in
      let b1 =
        TableBinding.put xm (ModStage m', FieldSet.cardinal fs) b in
      TableBinding.put xn (EgressStage n, x_size) b1
      in
    ([(nf_of_pred Id, TMods (xm, xmt, []))], egress_table (xn, xnt), theta')

  (* pipeline s (if b then p1 else p2) m *)
  | IfThen (a,p,q), m ->
    let metadata_p, metadata_q = split_metadata metadata estimate_mc_metadata p q in
    let m1, n1, theta1 = pipeline s metadata_p p m in
    let m2, n2, theta2 = pipeline s metadata_q q m in
    let m1' = List.map (fun (b,x) -> (nf_and a b, x)) m1 in
    let m2' = List.map (fun (b,x) -> (nf_and ((<!>) a) b, x)) m2 in
    let n = egress_stage_concat [qualify a n1; qualify ((<!>) a) n2] in
    (List.concat [m1'; m2'], n, compose theta1 theta2)

  (* pipeline s (p1 + p2) m *)
  | Union (p, q), m ->
    let f = try
        FieldSet.choose metadata
      with 
        Not_found -> raise (Failure "pipeline: no more metadata")
      in
    let metadata' = FieldSet.remove f metadata in
    let metadata_p, metadata_q = split_metadata metadata' estimate_mc_metadata p q in
    let m1, n1, theta1 = pipeline s metadata_p p m in
    let m2, n2, theta2 = pipeline s metadata_q q m in
    let m1' = List.map (mod_stage_add_mod f 0) m1 in
    let m2' = List.map (mod_stage_add_mod f 1) m2 in
    let m' = mod_stage_concat [m1'; m2'] in
    let n1' = qualify (Match (f,0)) n1 in
    let n2' = qualify (Match (f,1)) n2 in
    let n' = egress_stage_concat [n1'; n2'] in
    (m', n', compose theta2 theta1)

  (* pipeline s (p1; p2) m *)
  | Seq (p, q), m ->
    let metadata_p, metadata_q = split_metadata metadata estimate_mc_metadata p q in
    let m2, n2, theta2 = pipeline s metadata_q q m in
    let m1, n1, theta1 = pipeline s metadata_p p m2 in
    (m1, egress_stage_concat [n2; n1], compose theta2 theta1)

  (* pipeline s (p1 || p2) m *)
  | Con (p, _, _, q), m -> pipeline s metadata (Seq (p,q)) m


(** FIELD EXTRACTION *********************************************************)

(* Estimate the number of metadata fields necessary for field
 * extraction, not including tables. *)
let rec estimate_fe_metadata (p : pol) : int =
  match p with
  | Pred _
  | Modify _ 
  | Table _ -> 0
  | Union _ ->
    let msg = Printf.sprintf "(estimate_fe_metadata) union: (%s)" (print_pol p) in
    raise (Undefined msg)
  | Seq (p1, p2)
  | Con (p1, _, _, p2) -> (estimate_fe_metadata p1) + (estimate_fe_metadata p2)
  | IfThen (_, p1, p2) ->
    (estimate_fe_metadata p1) + (estimate_fe_metadata p2) + 1

let rec ext
  (f : Field.t) 
  (s : ((TableName.t * pol_t) -> FieldSet.t))
  (metadata : FieldSet.t)
  (pol : pol)
  (e : mod_stage) : (mod_stage * pol * theta) =
  match pol, e with

  (* ext f _ p drop *)
  | _, [] -> ([], Pred Id, (fun x -> x))

  (* ext f _ id e *)
  | Pred Id, _ -> (e, Pred Id, (fun x -> x))

  (* ext f drop e *)
  | Pred Drop, _ -> ([], Pred Drop, (fun x -> x))

  (* ext f s (f' = v) (e + a;m) *)
  | Pred a, _ ->
    let ra = infer_pred a in 
    if FieldSet.mem f ra then
      let e' = List.map (fun (b,m) -> (nf_and a b, m)) e in
      (e', Pred Id, (fun x -> x))
    else
      (e, pol, (fun x -> x))

  | Modify (f', v), e ->
    let handle_mod (a, m) = match m with
      (* ext f (f' <- v) (e + a; q) *)
      | Mods ms ->
        let a' = specialize f' v (pred_of_nf a) in
        if pred_equiv (Match (f', v) <^> a') Drop then
          (None, None)
        else if f = f' then
          (Some (nf_of_pred a', (Mods (pol::ms))), None)
        else
          (Some (nf_of_pred a', (Mods ms)), None)

      (* ext f (f' <- v) (e : a; q; (x : t)) *)
      | TMods (x, t, ms) ->
        let t1 = infer_pol (mod_stage_to_pol [a, (TMods (x, t, ms))]) in
        let x1 = TableName.gen () in
        let tx1 = pol_t_union t1 (FieldSet.empty, FieldSet.singleton f) in
        let theta1 = (fun b ->
          let _, x_size = TableBinding.get x b in
          let sub_ehead = sub_mods b [(a, (TMods (x, t, ms)))] in
          let e2, _, _ = ext f s metadata pol sub_ehead in
          TableBinding.put x1 (ModStage e2, x_size) b
        ) in
        if f = f' then
          (Some (nf_of_pred Id, TMods (x1, tx1, [])), Some theta1)
        else
          (Some (nf_of_pred Id, TMods (x1, tx1, [])), Some theta1)
      in
    let eopt, thetas = List.split (List.map handle_mod e) in
    let e' = List.filter Option.is_some eopt |> List.map Option.get in
    let theta (b : table_binding) = 
      List.filter Option.is_some thetas
      |> List.map Option.get
      |> List.fold_left (|>) b in
    let pol' = if f = f' then Pred Id else Modify (f', v) in
    (e', pol', theta)
  
  (* ext f s (x : (r,w)) e *)
  | Table (x, (r,w)), _ ->
    let fs = s (x, (r,w)) in
    let t = infer_pol (mod_stage_to_pol e) in
    let x1 = TableName.gen () in
    let x2 = TableName.gen () in
    let tx1 = pol_t_union (r, FieldSet.add f fs) t in
    let tx2 = (FieldSet.union r fs, FieldSet.remove f w) in
    let theta = (fun b ->
      let _, x_size = TableBinding.get x b in
      let e_size = int_of_float (2.0 ** (float_of_int (FieldSet.cardinal fs))) in
      let x_sub = sub_pol_forced b pol in
      let e_sub = sub_mods b e in
      let e1, p1, _ = ext f s fs x_sub e_sub in
      TableBinding.put x1 (ModStage e1, e_size) b |>
      TableBinding.put x2 (Policy p1, x_size)
    ) in
    ([nf_of_pred Id, TMods (x1, tx1, [])], Table (x2, tx2), theta)

  (* ext f s (if b then p else q) e *)
  | IfThen (b,p,q), _ ->
    let f' = try
        FieldSet.choose metadata
      with 
        Not_found -> raise (Failure "ext: no more metadata")
      in
    let metadata' = FieldSet.remove f' metadata in
    let metadata_p, metadata_q = split_metadata metadata' estimate_fe_metadata p q in
    let e1, p1, theta1 = ext f s metadata_p p e in
    let e2, p2, theta2 = ext f s metadata_q q e in
    let update_e a' v (a,x) =
      match x with
      | Mods ms -> (nf_and a' a, Mods ((Modify (f', v))::ms))
      | TMods (x,t,ms) -> (nf_and a' a, TMods (x, t, (Modify (f', v))::ms))
      in
    let e1' = List.map (update_e b 1) e1 in
    let e2' = List.map (update_e (Neg b) 0) e2 in
    ( mod_stage_concat [e1'; e2']
    , mkIf (Match (f',1)) p1 p2
    , compose theta2 theta1)

  (* union is not defined. *)
  | Union _, _ -> raise (Undefined "ext: not defined for union")

  (* ext f s (p; q) e *)
  | Seq (p, q), e ->
    let metadata_p, metadata_q = split_metadata metadata estimate_fe_metadata p q in
    let eq, q', thetaq = ext f s metadata_q q e in
    let ep, p', thetap = ext f s metadata_p p eq in
    (ep, p' <.> q', compose thetap thetaq)

  (* ext f s (p || q) e *)
  | Con (p, _, _, q), e -> ext f s metadata (Seq (p, q)) e


(** COMBINING MULTICAST CONSOLIDATION AND FIELD EXTRACTION *******************)

(* Combine the multicast consolidation and field extraction stages.  It is
 * assumed that the OutPort field was extracted. *)
let combine_mc_and_fe
  (m : mod_stage)
  (e : mod_stage) : (pol * pol * pol * theta) =
  let f = "OutPort" in
  (* Each summand in the extraction stage must have at least
   * one modification to OutPort.  Extract it.
   * Some summands, like tables, will not have this modification.
   * Collect those separately. *)
  let extract_f_mod
    (f : field)
    ((a, x) : NF.t * mod_sequence)
    : ((NF.t * mod_sequence * int), (NF.t * mod_sequence)) either =
    let fmods, non_fmods =
      List.partition
        (fun md -> match md with
          | Modify (f',_) -> f = f'
          | _ ->
            raise (Undefined ("(combine_mc_and_fe) field extraction " ^
                             "contains non-field modification policy")))
        (match x with Mods ps -> ps | TMods (_,_,ps) -> ps) in

    (* Tag pairs without modifications to f. *)
    if List.length fmods = 0 then Left (a,x) else

    (* Otherwise, count the last modification. *)
    let last_fmod = List.hd (List.rev fmods) in
    let last_fmod_v = match last_fmod with
      | Modify (_, v) -> v
      | _ ->
        raise (Undefined ("(combine_mc_and_fe) field extraction " ^
                         "contains non-field modification policy")) in
    match x with
    | Mods _ -> Right (a, Mods non_fmods, last_fmod_v)
    | TMods (x, t, _) -> Right (a, TMods (x, t, non_fmods), last_fmod_v)
    in

  (* Preemptively specialize the predicates. *)
  let e_spec = List.rev_map (fun (a,x) -> (nf_of_pred (specialize f 0 (pred_of_nf a)), x)) e in
  let m_spec = List.rev_map (fun (a,x) -> (nf_of_pred (specialize f 0 (pred_of_nf a)), x)) m in

  (* Remove elements with predicates equivalent to drop. *)
  let e_spec = List.filter (fun (a, _) -> not (pred_equiv (pred_of_nf a) Drop)) e_spec in
  let m_spec = List.filter (fun (a, _) -> not (pred_equiv (pred_of_nf a) Drop)) m_spec in

  (* Wrap everything else up in theta, to be done at population time. *)
  let x_stage_1 = TableName.gen () in
  let x_stage_2 = TableName.gen () in
  let x_stage_3 = TableName.gen () in

  let ts12 =
    pol_t_union 
      (infer_pol (mod_stage_to_pol e_spec))
      (FieldSet.empty, MetadataField.special_metadata) in
  let ts3 =
    pol_t_union 
      (infer_pol (mod_stage_to_pol e_spec))
      (FieldSet.empty, MetadataField.special_metadata) |>
      (pol_t_union (infer_pol (mod_stage_to_pol m_spec))) in

  let s1 = Table (x_stage_1, ts12) in
  let s2 = Table (x_stage_2, ts12) in
  let s3 = Table (x_stage_3, ts3) in

  let theta (b : table_binding) =

    let s1_size_estimate = height (mod_stage_to_pol e_spec) b in
    let s3_size_estimate = s1_size_estimate * (height (mod_stage_to_pol m_spec) b) in
    let m_sub = sub_mods b m_spec in
    let e_sub = sub_mods b e_spec in

    (* Group the extraction summation by OutPort. *)
    let e_ext_some_f = List.rev_map (extract_f_mod f) e_sub in
    let e_ext_f, ext_no_f =
      List.fold_left
        (fun (with_f, without_f) x ->
          match x with
          | Right y -> (y::with_f, without_f)
          | Left z -> (with_f, z::without_f))
        ([], [])
        e_ext_some_f in

    let e_groupby_v =
      List.fold_left
        (fun acc (a, x, v) ->
          try
            let mods = IntMap.find v acc in
            IntMap.add v ((a,x)::mods) acc
          with
            Not_found -> IntMap.add v [(a,x)] acc)
        IntMap.empty
        e_ext_f
      in
   
    (* Sanity check: field extraction summands without tables that don't modify
     * the field (no_mod_fe), when combined with the modification stage, should be
     * equivalent to drop. *)
    let () =
      list_cross m_sub ext_no_f |>
      List.iter (fun ((a, _), (b, _)) ->
        if pred_equiv (pred_of_nf a <^> pred_of_nf b) Drop then
          ()
        else
          ())
          (* TODO: revisit this. *)
          (* raise (Undefined "(combine_mc_and_fe) does not set OutPort.")) *)
      in

    (* ((tag = n; port = m), (a, mod_sequence)) *)
    let (e_tagged : ((int * int) * (NF.t * mod_sequence)) list) = 
      List.flatten
        (IntMap.fold
          (fun port (xs : (NF.t * mod_sequence) list) acc -> 
            let len_xs = List.length xs in
            let (prefixes : (int * int) list) =
              List.map
                (fun n -> (n, port))
                (1--len_xs) in
            let res =
              List.combine prefixes xs in
            res::acc)
        e_groupby_v
        [])
      in

    let (me_pairs :
      ((NF.t * mod_sequence) * ((int * int) * (NF.t * mod_sequence))) list) = 
      list_cross m_sub e_tagged in

    let (me_combined : ((int * int) * (pred * pol)) list) =
      List.rev_map
        (fun ((a, mmods), ((tag, port), (b, emods))) ->
          let test = nf_of_pred (Match ("tag", tag) <^> Match ("OutPort", port)) in
          let mpol = mod_sequence_to_pol mmods in
          let epol = mod_sequence_to_pol emods in
          let b' = specialize_across_mods (pred_of_nf b) mmods in
          let test' = 
            normalize (pred_of_nf test <^> pred_of_nf a <^> b') in
          ((tag, port), (test', mpol <.> epol)))
        me_pairs in
  
    (* Build the first stage by setting special metadata fields out_i. *)
    let (stage_1 : pol) =
      IntMap.fold
        (fun port _ acc ->
          let out_i = Printf.sprintf "out_%d" port in
          (Modify (out_i, 1)) <.> acc)
        e_groupby_v
        (Pred Id) in
  
    (* Build the second stage by matching on out_i, setting out <- i, and
     * setting a unique tag for each element in the groupby list. *)
    let tmp = List.fold_left
      (fun acc ((tag, port), (test, _)) ->
        if test = Drop then
          acc
        else
          IntPairSet.add (tag, port) acc)
      IntPairSet.empty
      me_combined in
    let (stage_2 : pol) =
      IntPairSet.fold
        (fun (tag, port) acc ->
          let out_i = Printf.sprintf "out_%d" port in
          let summand =
            Pred (Match (out_i, 1))
            <.> Modify ("tag", tag)
            <.> Modify ("OutPort", port) in
          summand <+> acc)
        tmp
        (Pred Drop) in

     let stage_3 =
      List.fold_left
        (fun acc (_, (test, t_branch)) -> mkIf test t_branch acc)
        (Pred Drop)
        me_combined in

    (* Build the table binding. *)
    TableBinding.put x_stage_1 (Policy stage_1, s1_size_estimate) b |>
      TableBinding.put x_stage_2 (Policy stage_2, s1_size_estimate) |>
      TableBinding.put x_stage_3 (Policy stage_3, s3_size_estimate)
    in
  (s1, s2, s3, theta)


(** TABLE FITTING ************************************************************)

let replace_with_table (p : pol) : pol * theta =
  let x = TableName.gen () in
  let tx = infer_pol p in
  let theta (b : table_binding) =
    let size_of_p = height p b in
    TableBinding.put x (Policy (sub_pol_forced b p), size_of_p) b in
  (Table (x, tx), theta)

let rec insert_tables (p : pol) : pol * theta =
  match p with
  | Pred _
  | Modify _
  | IfThen _ -> replace_with_table p
  | Union _ -> raise (Failure "(insert_tables) union")
  | Seq (p1, p2) ->
    let p1', theta1 = insert_tables p1 in
    let p2', theta2 = insert_tables p2 in
    (p1' <.> p2', compose theta2 theta1)
  | Con (p1, w1, w2, p2) ->
    let p1', theta1 = insert_tables p1 in
    let p2', theta2 = insert_tables p2 in
    (Con (p1', w1, w2, p2'), compose theta2 theta1)
  | Table _ -> (p, (fun x -> x))

(* Number of bits p can use in matches. *)
let width (p : pol) : int =
  let mkFieldMap (fs : (Field.t * int) list) : int FieldMap.t =
    List.fold_left (fun acc (f,v) -> FieldMap.add f v acc) FieldMap.empty fs in
  let field_costs = mkFieldMap (List.concat [
    [ ("SrcMAC", 48)
    ; ("DstMAC", 48)
    ; ("SrcIP",  32)
    ; ("DstIP",  32)
    ; ("EthTyp", 16)
    ; ("VLAN",   16)
    ; ("InPort",  6)
    ; ("OutPort", 6)
    ];
    List.map
      (fun f -> (f, 1))
      (FieldSet.elements MetadataField.special_metadata) ]) in
  let (r,_) = infer_pol p in
  let r_width =
    List.fold_left (+) 0
      (List.map
        (fun k -> try
            FieldMap.find k field_costs
          with _ ->
            (* Metadata fields are one bit. *)
            1)
        (FieldSet.elements r)) in
  r_width

let blocks
  (p : pol)
  (table_estimates : table_binding) : int =
  let w = ceil ((float_of_int (width p)) /. 40.0) in
  let h = ceil ((float_of_int (height p table_estimates)) /. 2048.0) in
  (max 1 (int_of_float w)) * (int_of_float h)

module IntPairMap = Map.Make( struct
  type t = int * int
  let compare = compare
end )

type cost =
  | Inf
  | Cost of int

let print_cost (c : cost) : string = match c with
  | Inf -> "Inf"
  | Cost n -> string_of_int n

let (|+) (n : cost) (m : cost) : cost = match n, m with
  | Inf, _ -> Inf
  | _, Inf -> Inf
  | Cost v1, Cost v2 -> Cost (v1 + v2)

let cost_min (n : cost) (m : cost) : cost = match n, m with
  | Inf, _ -> m
  | _, Inf -> n
  | Cost v1, Cost v2 -> Cost (min v1 v2)

let (|<) (n : cost) (m : cost) : bool = match n, m with
  | Inf, Inf -> false
  | Inf, _ -> false
  | _, Inf -> true
  | Cost v1, Cost v2 -> v1 < v2

let array_equal
  (f : 'a -> 'a -> bool)
  (arr1 : 'a array)
  (arr2 : 'a array) : bool =
  try
    let a12 = List.combine (Array.to_list arr1) (Array.to_list arr2) in
    List.fold_left (fun acc (x,y) -> acc && f x y) true a12
  with
    Invalid_argument _ -> false

let matrix_equal
  (f : 'a -> 'a -> bool)
  (m1 : 'a array array)
  (m2 : 'a array array) : bool =
  array_equal (array_equal f) m1 m2

let print_matrix (f : 'a -> string) (arr : 'a array array) : string =
  let print_row (row : 'a array) : string =
    string_join " | " (List.map f (Array.to_list row)) in
  string_join "\n" (List.map print_row (Array.to_list arr))

let matrix_to_list (m : 'a array array) : 'a list list =
  List.map Array.to_list (Array.to_list m)

module Array3 = struct
  
  type 'a t = 'a array array array

  let equal
    (f : 'a -> 'a -> bool)
    (m1 : 'a t)
    (m2 : 'a t) : bool =
    array_equal (matrix_equal f) m1 m2

  let make n1 n2 n3 (v : 'a) : 'a t =
    (* Create a 2D array with throw-away elements, because make_matrix 
     * shares the initializer across all elements. *)
    let arr = Array.make_matrix n1 n2 (Array.make 0 v) in
    (* Initialize 2D array with distinct elements. *)
    let () = 
      for i = 0 to (n1 - 1) do
        for j = 0 to (n2 - 1) do
          arr.(i).(j) <- (Array.make n3 v)
        done
      done in
    arr

  let print (f : 'a -> string) (x : 'a t) : string =
    print_matrix 
      (fun arr -> string_join ", " (List.map f (Array.to_list arr)))
      x

  let to_list (x : 'a t) : 'a list list list =
    List.map matrix_to_list (Array.to_list x)

end

(* CostArray(x,y,z).
 *   x: starting index of subarray.
 *   y: ending index of subarray (inclusive).
 *   z: number of partitions allowed in tx..ty.
 *)
module CostArray = struct
  type range = cost
  type t = range array array
  let get ((n, m) : int * int) (x : t) : range = (x).(n).(m)
  let set ((n, m) : int * int)
          (v : range)
          (x : t) : unit = (x).(n).(m) <- v
  let make n : t = Array.make_matrix n n Inf
  let print (x : t) : string = print_matrix print_cost x
  let to_list (x : t) : range list list = matrix_to_list x
  let equal (x : t) (y : t) : bool = matrix_equal (=)  x y
end

let (|/) (m : int) (n : int) : int =
  (float_of_int m) /. (float_of_int n) |> ceil |> int_of_float

(* m, s = build_cost_tables xs table_estimates
 * m[i,j] = smallest number of tables xi..xj fit into.
 * s[i,j] = index at which xi..xj is partitioned, if any.
 *)
let build_cost_tables
  (xs : (TableName.t * pol_t) list)
  (table_estimates : table_binding)
  : CostArray.t * int option array array =
  (* Unlike the algorithm on paper, indices begin at zero. *)
  let n = List.length xs in
  let xarr = Array.of_list xs in
  let m = CostArray.make n in
  let s = Array.make_matrix n n None in

  (* Wrap the blocks function for convenience. *)
  let blocks_of_arr (xarr : (TableName.t * pol_t) array) : int =
    let pol_of_arr =
      Array.fold_left (fun acc (x,t) -> Table (x,t) <.> acc) (Pred Id) xarr in
    blocks pol_of_arr table_estimates
    in

  (* Initialize m. *)
  for i = 0 to (n-1) do
    CostArray.set (i,i) (Cost (blocks_of_arr (Array.sub xarr i 1) |/ 16)) m
  done;

  (* Analyze all list subsets. *)
  for len = 1 to n do
    for i = 0 to (n - len) do
      let j = i + len - 1 in
      for k = i to (j - 1) do
        let partition_cost =
          (CostArray.get (i,k) m) |+ (CostArray.get (k+1, j) m) in
        let one_big_table_cost = Cost
          (blocks_of_arr (Array.sub xarr i (j - i + 1)) |/ 16) in
        let q = cost_min partition_cost one_big_table_cost in
        let k_recorded =
          if partition_cost |< one_big_table_cost then
            Some (k - i)
          else
            None
          in
        (if q |< (CostArray.get (i,j) m) then
          (CostArray.set (i,j) q m;
           s.(i).(j) <- k_recorded)
        else
          ())
      done
    done
  done;

  (m, s)

module TableListSet = Set.Make( struct
  type t = (TableName.t * pol_t) list
  let compare = Pervasives.compare
end )

(* Build all permutations of xs. *)
let build_xs p : (TableName.t * pol_t) list list = 
  let rec aux p =
    match p with
    | Pred _
    | Modify _
    | IfThen _ -> raise (Failure "(partition) found non-table primitive")
    | Table (x,t) -> TableListSet.singleton [(x,t)]
    | Union _ -> raise (Failure "(partition) found union")
    | Seq (p1, p2) ->
      let xs1 = aux p1 in
      let xs2 = aux p2 in 
      let crossed =
        list_cross
          (TableListSet.elements xs1)
          (TableListSet.elements xs2) in
      let combined =
        List.map (fun (xs1, xs2) -> List.concat [xs1; xs2]) crossed in
      List.fold_left
        (fun acc x -> TableListSet.add x acc)
        TableListSet.empty
        combined
    | Con (p1, _, _, p2) ->
      let xs1 = aux p1 in
      let xs2 = aux p2 in 
      let crossed =
        list_cross
          (TableListSet.elements xs1)
          (TableListSet.elements xs2) in
      let combined_with_permutations =
          List.map
            (fun (xs1, xs2) ->
              [List.concat [xs1; xs2]; List.concat [xs2; xs1]])
            crossed in
      let combined = List.flatten combined_with_permutations in
      List.fold_left
        (fun acc x -> TableListSet.add x acc)
        TableListSet.empty
        combined
  in
  TableListSet.elements (aux p)


(* Return a partitioning of xs, along with the number of physical
 * tables required for each partition---i.e. when one table variable
 * is too large for one physical table.
 *
 * Each list element (v, tables) is a single "logical" table.
 *)
let partition_internal
  (xs : (TableName.t * pol_t) list)
  (table_estimates : table_binding)
  : (int * ((TableName.t * pol_t) list)) list =
  let m, s = build_cost_tables xs table_estimates in
  let xarr = Array.of_list xs in
  let rec aux i j =
    match s.(i).(j), (CostArray.get (i,j) m) with
    | None, Cost v -> [(v, Array.to_list (Array.sub xarr i (j - i + 1)))]
    | Some partition, _ ->
      List.concat [(aux i partition); (aux (partition + 1) j)]
    | None, Inf -> raise (Failure "(partition) infinite cost")
    in
  aux 0 ((List.length xs) - 1)


(* Return a partitioning of p, along with the number of physical
 * tables required for each partition---i.e. when one table variable
 * is too large for one physical table.
 *)
let partition 
  (p : pol)
  (table_estimates : table_binding)
  : (int * ((TableName.t * pol_t) list)) list =
  let cost_of_partition part =
    List.fold_left (fun acc (v, _) -> v + acc) 0 part in
  let (list_of_xs : (TableName.t * pol_t) list list) = build_xs p in
  let partitioned_xs =
    List.map
      (fun xs -> partition_internal xs table_estimates)
      list_of_xs in
  let opt_best = List.fold_left
    (fun opt part -> match opt with
      | Some best_part ->
        let part_cost = cost_of_partition part in
        let best_part_cost = cost_of_partition best_part in
        if part_cost < best_part_cost then
          Some part
        else
          Some best_part
      | None -> Some part)
    None
    partitioned_xs in
  match opt_best with
  | Some best -> best
  | None -> raise (Failure "(partition) no results")

let size_of_partition
  (partition : (int * ((TableName.t * pol_t) list)) list) : int =
  List.fold_left (fun acc (v,_) -> acc + v) 0 partition
