(*
 * The MIT License (MIT)
 * 
 * Copyright (c) 2015 Cole Schlesinger
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *)

open My_core
open Syntax

(* Equivalence decision procedure for predicates. *)

(* Predicate normal form: a summation of sequences of matches and negations of
 * single matches. *)

(* Pair of positive/negative matches.  A pair of empty maps indicates 'id'. *)
module NF_Seq = struct
  type t = (int FieldMap.t) * (IntSet.t FieldMap.t)
  let compare (p1, n1) (p2, n2) =
    let res = FieldMap.compare compare p1 p2 in
    if res = 0 then
      FieldMap.compare compare n1 n2
    else
      res
end

(* Concatenate two NF_Seqs.  None indicates the concatenated sequence is
 * equivalent to drop. *)
let nf_seq_concat ((p1,n1) : NF_Seq.t) ((p2,n2) : NF_Seq.t) : NF_Seq.t option =
  try
    let p12 = FieldMap.merge
      (fun _ a b -> match a,b with
        | Some v1, Some v2 ->
          if v1 = v2 then Some v1 else raise (Undefined "")
        | Some v1, _ -> Some v1
        | _, Some v2 -> Some v2
        | None, None -> None)
      p1 p2 in
    let n12 = FieldMap.merge
      (fun _ a b -> match a,b with
        | Some s1, Some s2 -> Some (IntSet.union s1 s2)
        | Some s1, _ -> Some s1
        | _, Some s2 -> Some s2
        | None, None -> None)
      n1 n2 in
    (* Use merge to check whether the same field/value is in p12 and n12. *)
    let _ = FieldMap.merge
      (fun _ a b -> match a,b with
        | Some v, Some s ->
          if IntSet.mem v s then raise (Undefined "") else None
        | _ -> None)
      p12 n12 in
    Some (p12, n12)
  with
    Undefined _ -> None

(* Sum of sequences.  An empty set indicates 'drop'. *)
module NF = Set.Make(NF_Seq)

let mkNF (nfs : NF_Seq.t list) : NF.t =
  List.fold_left
    (fun acc e -> NF.add e acc)
    NF.empty
    nfs

(* Cross product of two sums. *)
let nf_cross (nf1 : NF.t) (nf2 : NF.t) : NF.t =
  let l1 = NF.elements nf1 in
  let l2 = NF.elements nf2 in
  let (paired : (NF_Seq.t * NF_Seq.t) list) = List.flatten 
    (List.map
      (fun s1 -> List.map (fun s2 -> (s1, s2)) l2)
      l1) in
  let opt_nf_seqs = List.map (fun (x,y) -> nf_seq_concat x y) paired in
  List.fold_left
    (fun acc nfopt ->
      match nfopt with
      | Some nf_seq -> NF.add nf_seq acc
      | None -> acc)
    NF.empty
    opt_nf_seqs

(* Negate a sequence: a; b; -c; -d == -a + -b + c + d. *)
let nf_seq_negate ((pmap, nmap) : NF_Seq.t) : NF.t =
  (* Build negation of pmap. *)
  let ps = FieldMap.bindings pmap in
  let (neg_ps : NF_Seq.t list) = 
    List.map
      (fun (f,v) ->
        (FieldMap.empty, FieldMap.singleton f (IntSet.singleton v)))
      ps in
  let neg_ps_nf = mkNF neg_ps in

  (* Build negation of nmap. *)
  let ns = List.flatten
    (List.map 
      (fun (f, s) ->
        List.map (fun v -> FieldMap.singleton f v) (IntSet.elements s))
      (FieldMap.bindings nmap)) in
  let neg_ns = List.map (fun x -> (x, FieldMap.empty)) ns in
  let neg_ns_nf = mkNF neg_ns in

  (* Join. *)
  NF.union neg_ps_nf neg_ns_nf

(* NF constants. *)
let nf_id = NF.singleton (FieldMap.empty, FieldMap.empty)
let nf_drop = NF.empty

(* Negate a sum of sequences. *)
let rec nf_negate (nf : NF.t) : NF.t =
  try
    let s = NF.choose nf in
    let not_s = nf_seq_negate s in
    let the_rest = nf_negate (NF.remove s nf) in
    nf_cross not_s the_rest
  with
    Not_found -> nf_id

let rec nf_of_pred (a : pred) : NF.t =
  match a with
  | Id -> nf_id
  | Drop -> nf_drop
  | Match (f,v) -> NF.singleton (FieldMap.singleton f v, FieldMap.empty)
  | Neg b ->
    let nfb = nf_of_pred b in
    nf_negate nfb
  | And (b1, b2) ->
    let nfb1 = nf_of_pred b1 in
    let nfb2 = nf_of_pred b2 in
    nf_cross nfb1 nfb2
  | Or (b1, b2) -> 
    let nfb1 = nf_of_pred b1 in
    let nfb2 = nf_of_pred b2 in
    NF.union nfb1 nfb2

let nf_and (a : pred) (nf : NF.t) : NF.t =
  let a_nf = nf_of_pred a in
  nf_cross a_nf nf

let pred_of_nf_seq ((pos, neg) : NF_Seq.t) : pred =
  let pos_pr =
    FieldMap.fold
      (fun f v acc -> Match (f, v) <^> acc)
      pos Id in
  let neg_pr = FieldMap.fold
    (fun f vs acc ->
      if FieldMap.mem f pos then
        acc
      else
        (IntSet.fold
          (fun v acc' ->
            Neg (Match (f,v)) <^> acc')
          vs Id) <^> acc)
    neg
    Id
    in
  pos_pr <^> neg_pr

let pred_of_nf (nf : NF.t) : pred =
  NF.fold (fun n acc -> (pred_of_nf_seq n) <||> acc) nf Drop

let print_nf (nf : NF.t) : string = pred_of_nf nf |> print_pred

let pred_equiv (a : pred) (b : pred) : bool =
  let nfa = nf_of_pred a in
  let nfb = nf_of_pred b in
  NF.equal nfa nfb

let normalize (a : pred) : pred = nf_of_pred a |> pred_of_nf
