(*
 * The MIT License (MIT)
 * 
 * Copyright (c) 2015 Cole Schlesinger
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *)

module Result = struct
  type ('a, 'b) t =
    | Ok of 'a
    | Error of 'b
  let (>>=) (res : ('a, 'b) t) (f : 'a -> ('c, 'b) t) : ('c, 'b) t =
    match res with
    | Ok v -> f v
    | Error e -> Error e
  let all (results : ('a, 'b) t list) : ('a list, 'b) t =
    let rec aux (acc : ('a list, 'b) t) (rs : ('a, 'b) t list)
      : ('a list, 'b) t =
      match rs with
      | [] -> acc
      | (Ok v)::tail -> begin match aux acc tail with
        | Error _ -> acc
        | Ok vs -> Ok (v::vs)
      end
      | (Error e)::_ -> Error e
      in
    match aux (Ok []) results with
    | Ok res_rev -> Ok (List.rev res_rev)
    | Error e -> Error e
end

let (>>=) = Result.(>>=)

(* Useful types. *)
type ('a, 'b) either =
  | Right of 'a
  | Left of 'b

(* Helper exceptions. *)

(* Indicates a function is undefined on the given input. *)
(* TODO: unify this with other indications of undefinedness, like Result.t. *)
exception Undefined of string

(* Helper functions. *)
let string_join (joiner : string) (l : string list) =
  let rec f l' =
    match l' with
    | [] -> ""
    | [x] -> x
    | h::ts -> h ^ joiner ^ (f ts) in
  Printf.sprintf "[%s]" (f l)

let print_str_list (l : string list) = string_join ", " l

let rec take (n : int) (l : 'a list) : ('a list * 'a list) =
  if n <= 0 then
    ([], l)
  else
  match l with
  | [] -> raise (Failure "take")
  | h::t ->
    let taken, remainder = take (n-1) t in
    (h::taken, remainder)

let list_cross (l1 : 'a list) (l2 : 'b list) : (('a * 'b) list) =
  List.flatten
    ((List.map (fun a ->
     List.map (fun b -> (a, b)) l2)) l1)

let (--) (i : int) (j : int) =
  let rec aux n acc =
    if n < i then acc else aux (n-1) (n :: acc)
    in
  aux j []

(* Useful modules. *)
module IntSet = Set.Make(
  struct
    type t = int
    let compare = Pervasives.compare
  end )

module IntMap = Map.Make(
  struct
    type t = int
    let compare = Pervasives.compare
  end )

module IntPairSet = Set.Make(
  struct
    type t = int * int
    let compare = Pervasives.compare
  end )

module Option = struct
  type 'a t = 'a option
  let is_some x : bool = match x with
    | Some _ -> true
    | None -> false
  let get (x : 'a option) : 'a =  match x with
    | Some v -> v
    | None -> raise Not_found
end
