(*
 * The MIT License (MIT)
 * 
 * Copyright (c) 2015 Cole Schlesinger
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *)

open My_core
open Syntax

(* Packet trees. *)
type packet_tree =
  | Leaf of (PacketSet.t * FieldSet.t)
  | ParNode of (packet_tree * packet_tree)
  | NegNode of (PacketSet.t * packet_tree)
  | ConNode of (FieldSet.t * packet_tree * packet_tree)
  | SeqNode of packet_tree

let rec print_packet_tree (tree : packet_tree) : string =
  match tree with
  | Leaf (pks, w) ->
    Printf.sprintf "<%s, %s>" (print_packets pks) (print_fields w)
  | ParNode (tr1, tr2) ->
    let s1 = (print_packet_tree tr1) in
    let s2 = (print_packet_tree tr2) in
    Printf.sprintf "Par (%s) (%s)" s1 s2
  | NegNode (pks, tr) ->
    let pks_s = print_packets pks in
    let tr_s = print_packet_tree tr in
    Printf.sprintf "Neg (%s) (%s)" pks_s tr_s
  | ConNode (w, tr1, tr2) ->
    let w_s = print_fields w in
    let tr1_s = print_packet_tree tr1 in
    let tr2_s = print_packet_tree tr2 in
    Printf.sprintf "Con (%s) (%s) (%s)" w_s tr1_s tr2_s
  | SeqNode t1 ->
    let s1 = print_packet_tree t1 in
    Printf.sprintf "Seq (%s)" s1

(* States. *)
type state = (pol * packet_tree)
type pr_state = (pred * packet_tree)

let print_state ((p, tree) : state) : string =
  Printf.sprintf "<| %s, %s |>" (print_pol p) (print_packet_tree tree)


(* Table binding updates. *)


(* Packet cross product.  Fails if both packets have the same field with
 * different values. *)
(* TODO: refactor this monstrosity. *)
let cross
  (s1 : PacketSet.t) 
  (s2 : PacketSet.t) : (PacketSet.t, string) Result.t =
  let s1_list = PacketSet.elements s1 in
  let s2_list = PacketSet.elements s2 in
  let combine pk1 pk2 : (Packet.t, string) Result.t =
    let merge _ v1opt v2opt : ((int, unit) Result.t) option =
      match v1opt, v2opt with
      | Some v1, Some v2 ->
        if v1 = v2 then
          Some (Ok v1)
        else
          Some (Error ())
      | Some v, _
      | _, Some v -> Some (Ok v)
      | _, _ -> None
      in
    let merged : (int, unit) Result.t FieldMap.t =
      FieldMap.merge merge pk1 pk2 in
    let errors =
      FieldMap.filter
        (fun _ (e : (int, unit) Result.t) ->
          match e with Ok _ -> false | Error _ -> true)
        merged in
    if FieldMap.is_empty errors then
      let good_merged_pkt =
        FieldMap.map
          (fun (e : (int, unit) Result.t) ->
            match e with Ok v -> v | Error _ -> failwith "impossible")
          merged
        in
      Ok good_merged_pkt
    else
      let msg =
        Printf.sprintf
          "packet cross failed: (%s) X (%s)"
          (print_packet pk1)
          (print_packet pk2) in
      Error msg
    in
  let combined =
    List.flatten (List.map
      (fun pk1 -> 
        List.map (fun pk2 -> combine pk1 pk2)
        s2_list)
      s1_list) in
  let errors = 
    List.filter
      (fun (e : (Packet.t, string) Result.t) ->
        match e with Ok _ -> false | Error _ -> true)
      combined in
  if List.length errors = 0 then
    let combined_good =
      List.map
        (fun (e : (Packet.t, string) Result.t) ->
          match e with Ok v -> v | Error _ -> failwith "impossible")
        combined in
    Ok (mkPacketSet combined_good)
  else
    let msg =
      Printf.sprintf
        "packet cross failed: (%s) X (%s)"
        (print_packets s1)
        (print_packets s2) in
    Error msg

let remove_fields (wq : FieldSet.t) (pks : PacketSet.t) =
  let filter f _ = not (FieldSet.mem f wq) in
  let pk_list = List.map (FieldMap.filter filter) (PacketSet.elements pks) in
  mkPacketSet pk_list



(* Predicate single-step evaluation. *)
let rec step_pred ((pr, tree) : pr_state) : (pr_state, string) Result.t =
  match (pr, tree) with
  (* Id *)
  | Id, Leaf _ -> Ok (pr, tree)

  (* Drop *)
  | Drop, Leaf (_, w) -> Ok (Id, Leaf (PacketSet.empty, w))

  (* Match *)
  | Match (f, v), Leaf(pks, w) ->
    let filter pk = 
      try
        v = (FieldMap.find f pk)
      with
        _ -> false in
    let pks' = PacketSet.filter filter pks in
    Ok (Id, Leaf (pks', w))

  (* NotEnter *)
  | Neg pr, Leaf(pks, _) -> Ok (pr, (NegNode (pks, tree)))

  (* NotExit *)
  | Id, NegNode (pks, Leaf (pks_a, w)) ->
    Ok (Id, Leaf (PacketSet.diff pks pks_a, w))

  (* NotInner *)
  | a, NegNode (pks, tree') ->
    step_pred (a, tree') >>= fun (a', tree'') ->
    Ok (a', NegNode (pks, tree''))

  (* SeqEnter *)
  | And _, Leaf _ -> Ok (pr, SeqNode tree)

  (* SeqExitR *)
  | And (Id, b), SeqNode (Leaf (pks, w)) -> Ok (b, Leaf (pks, w))

  (* SeqL *)
  | And (a, b), SeqNode tree ->
    step_pred (a, tree) >>= fun (a', tree') ->
    Ok (And (a', b), SeqNode tree')

  (* ParExit *)
  | Or (Id, Id), ParNode (Leaf (pks1, w1), Leaf (pks2, w2)) ->
    if FieldSet.compare w1 w2 = 0 then
      Ok (Id, Leaf (PacketSet.union pks1 pks2, w1))
    else
      let msg =
        Printf.sprintf
          "ParExit with mismatched write sets: %s"
          (print_state (Pred pr, tree)) in
      Error msg

  (* ParEnter *)
  | Or _, Leaf _ -> Ok (pr, ParNode (tree, tree))

  (* ParR *)
  | Or (Id, b), ParNode (Leaf (pks, w), tree2) ->
    step_pred (b, tree2) >>= fun (b', tree2') ->
    Ok (Or (Id, b'), ParNode (Leaf (pks, w), tree2'))

  (* ParL *)
  | Or (a, b), ParNode (tree1, tree2) ->
    step_pred (a, tree1) >>= fun (a', tree1') ->
    Ok (Or (a', b), ParNode (tree1', tree2))

  (* Stuck *)
  | _, _ ->
    let msg =
      Printf.sprintf
        "invalid state: %s" 
        (print_state (Pred pr, tree)) in
    Error msg

(* Policy single-step evaluation. *)
let rec step_pol ((pol, tree) : state) : (state, string) Result.t =
  match pol, tree with
  (* Modify *)
  | Modify (f, v), Leaf (pks, w) ->
    if FieldSet.mem f w then
      let pks' =
        PacketSet.fold
          (fun pk acc ->
            let pk' = FieldMap.add f v pk in
            PacketSet.add pk' acc)
          pks
          PacketSet.empty in
      Ok (Pred Id, Leaf (pks', w))
    else
      let msg =
        Printf.sprintf
          "tried to modify a packet not in the write set: %s"
          (print_state (pol, tree)) in
      Error msg

  (* SeqEnter *)
  | Seq _, Leaf _ -> Ok (pol, SeqNode tree)

  (* SeqExitR *)
  | Seq (Pred Id, q), SeqNode (Leaf (pks, w)) -> Ok (q, Leaf (pks, w))

  (* SeqL *)
  | Seq (p, q), SeqNode tree ->
    step_pol (p, tree) >>= fun (p', tree') ->
    Ok (Seq (p', q), SeqNode tree')

  (* Desugar IfThen to Union. *)
  | IfThen (a, p, q), Leaf (pks, w) ->
    let desugared = desugar_if (a, p, q) in
    Ok (desugared, Leaf (pks, w))

  (* ParExit *)
  | Union (Pred Id, Pred Id), ParNode (Leaf (pks1, w1), Leaf (pks2, w2)) ->
    if FieldSet.compare w1 w2 = 0 then
      Ok (Pred Id, Leaf (PacketSet.union pks1 pks2, w1))
    else
      let msg =
        Printf.sprintf
          "ParExit with mismatched write sets: %s"
          (print_state (pol, tree)) in
      Error msg

  (* ParEnter *)
  | Union _, Leaf _ -> Ok (pol, ParNode (tree, tree))

  (* ParR *)
  | Union (Pred Id, q), ParNode (Leaf (pks, w), tree2) ->
    step_pol (q, tree2) >>= fun (q', tree2') ->
    Ok (Union (Pred Id, q'), ParNode (Leaf (pks, w), tree2'))

  (* ParL *)
  | Union (p, q), ParNode (tree1, tree2) ->
    step_pol (p, tree1) >>= fun (p', tree1') ->
    Ok (Union (p', q), ParNode (tree1', tree2))

  (* ConExit *)
  | Con (Pred Id, _, _, Pred Id),
    ConNode (w, Leaf (pkp, _), Leaf (pkq, _)) ->
    cross pkp pkq >>= fun pks ->
    Ok (Pred Id, Leaf (pks, w))

  (* ConEnter *)
  | Con (_, wp, wq, _), Leaf (pks, w) ->
    let ws_inter = FieldSet.inter wp wq in
    let is_sub = FieldSet.subset (FieldSet.union wp wq) w in
    if (FieldSet.is_empty ws_inter) && is_sub then
      let pks_p = remove_fields wq pks in
      let pks_q = remove_fields wp pks in
      Ok (pol, ConNode (w, Leaf (pks_p, wp), Leaf (pks_q, wq)))
    else
      let msg =
        Printf.sprintf
          "invalid state: (ConEnter got stuck) %s"
          (print_state (pol, tree)) in
      Error msg

  (* ConR *)
  | Con (Pred Id, wp, wq, q), ConNode (w, Leaf lp, tree) ->
    step_pol (q, tree) >>= fun (q', tree') ->
    Ok (Con (Pred Id, wp, wq, q'), ConNode (w, Leaf lp, tree'))

  (* ConL *)
  | Con (p, wp, wq, q), ConNode (w, tree_p, tree_q) ->
    step_pol (p, tree_p) >>= fun (p', tree_p') ->
    Ok (Con (p', wp, wq, q), ConNode (w, tree_p', tree_q))

  (* Pred *)
  | Pred pr, _ ->
    step_pred (pr, tree) >>= fun (pr', tree') ->
    Ok (Pred pr', tree')

  (* Stuck *)
  | _, _ ->
    let msg =
      Printf.sprintf
        "invalid state: %s" 
        (print_state (pol, tree)) in
    Error msg

(* Transitive closure of the step relation. *)
let rec step_many (st : state) : (PacketSet.t, string) Result.t =
  match st with
  | Pred Id, Leaf (pks, _) -> Ok pks
  | _, _ ->
    step_pol st >>= fun st' ->
    step_many st'

(* Policy evaluation. *)
let eval (p : pol) (pks : PacketSet.t) : (PacketSet.t, string) Result.t =
  let _, w = infer_pol p in
  let pks_list = PacketSet.elements pks in

  (* Evaluate each packet, one at a time. *)
  let (results : ((PacketSet.t, string) Result.t) list) =
    List.map
      (fun pk -> step_many (p, Leaf (PacketSet.singleton pk, w)))
      pks_list in

  Result.all results >>= fun (result_pks : PacketSet.t list) ->
  let pks_list' = List.flatten (List.map PacketSet.elements result_pks) in
  Ok (mkPacketSet pks_list')
  
