(*
 * The MIT License (MIT)
 * 
 * Copyright (c) 2015 Cole Schlesinger
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *)


open My_core

(*** Fields and packets. ***)

type field = string
let print_field f = f
module Field = struct
  type t = string
  let compare = Pervasives.compare
end

module FieldSet = Set.Make(Field)

let inter_is_empty (s1 : FieldSet.t) (s2 : FieldSet.t) : bool =
  FieldSet.is_empty (FieldSet.inter s1 s2)

let print_fields (fs : FieldSet.t) : string =
  Printf.sprintf "{%s}" (string_join ", " (FieldSet.elements fs))

let mkFieldSet (fs : Field.t list) : FieldSet.t =
  List.fold_left
    (fun acc f -> FieldSet.add f acc)
    FieldSet.empty
    fs

module FieldMap = Map.Make(Field)

module Packet = struct
  type t = int FieldMap.t
  let compare p1 p2 = FieldMap.compare compare p1 p2
  let empty : t = (FieldMap.empty : int FieldMap.t)
end
let print_packet (pk : Packet.t) : string =
  let str_list =
    List.map
      (fun (k,v) -> Printf.sprintf "(%s, %d)" (print_field k) v)
      (FieldMap.bindings pk) in
  print_str_list str_list

module PacketSet = Set.Make(Packet)
let print_packets (pks : PacketSet.t) : string =
  let packet_list = PacketSet.elements pks in
  let packet_str_list = List.map print_packet packet_list in
  print_str_list packet_str_list

let mkPacketSet (pks : Packet.t list) : PacketSet.t =
  List.fold_left
    (fun acc pk -> PacketSet.add pk acc)
    PacketSet.empty
    pks

(* Generate fresh table names. *)
module TableName : sig
    type t
    val gen : unit -> t
    val print : t -> string
    val compare : t -> t -> int
end = struct
    type t = string
    let count = ref 0
    let gen () =
        let rv = Printf.sprintf "t%d" !count in
        let () = count := !count + 1 in
        rv
    let print v = v
    let compare = Pervasives.compare
end

(* Generate fresh field names. *)
module MetadataField : sig
    type t
    val gen : unit -> t
    val gen_many : int -> FieldSet.t
    val print : t -> string
    val compare : t -> t -> int
    val special_metadata : FieldSet.t
end = struct
    type t = string
    let count = ref 0
    let gen () =
        let rv = Printf.sprintf "f%d" !count in
        let () = count := !count + 1 in
        rv
    let gen_many n =
      let rec f n acc =
        if n <= 0 then 
          acc
        else
          FieldSet.add (gen ()) (f (n - 1) acc)
        in
      f n FieldSet.empty
    let print v = v
    let compare = Pervasives.compare
    let special_metadata =
      let out_fields =
        List.map (fun n -> Printf.sprintf "out_%d" n) (-1--16) in
      mkFieldSet ("tag"::out_fields)
end



(* Policy types: ({READ}, {WRITE}). *)
type pol_t = FieldSet.t * FieldSet.t

let print_pol_t (r,w) =
  "(" ^ print_fields r ^ ", " ^ print_fields w ^ ")"
let pol_t_equiv ((r1,w1) : pol_t) ((r2,w2) : pol_t) : bool =
  (FieldSet.equal r1 r2) && (FieldSet.equal w1 w2)
let (<:) ((r1,w1) : pol_t) ((r2,w2) : pol_t) : bool =
  (FieldSet.subset r1 (FieldSet.union r2 w2)) && (FieldSet.subset w1 w2)
let pol_t_compare ((r1,w1) : pol_t) ((r2,w2) : pol_t) : int =
  let res = FieldSet.compare r1 r2 in
  if res = 0 then
    FieldSet.compare w1 w2
  else
    res

(* Pretty printer for the debugger. *)
let pp_pol_t ch p = Format.fprintf ch "%s" (print_pol_t p)

let pol_t_union ((r1,w1) : pol_t) ((r2,w2) : pol_t) : pol_t =
  (FieldSet.union r1 r2, FieldSet.union w1 w2)

module TableSet = Set.Make( struct
  type t = TableName.t
  let compare = Pervasives.compare
end )

type pred =
    | Id
    | Drop
    | Match of Field.t * int
    | Neg of pred
    | And of pred * pred
    | Or of pred * pred

let rec print_pred (pr : pred) : string =
  match pr with
  | Id -> "Id"
  | Drop -> "Drop"
  | Match (f, v) -> Printf.sprintf "%s = %d" (print_field f) v
  | Neg pr' -> Printf.sprintf "-(%s)" (print_pred pr')
  | And (pr1, pr2) ->
    let rv1 = print_pred pr1 in
    let rv2 = print_pred pr2 in
    rv1 ^ " ^ " ^ rv2
  | Or (pr1, pr2) ->
    let rv1 = print_pred pr1 in
    let rv2 = print_pred pr2 in
    rv1 ^ " || " ^ rv2

let rec size_of_pred (pr : pred) : int =
  match pr with
  | Id | Drop | Match _ -> 1
  | Neg a -> (size_of_pred a) + 1
  | And (a,b) | Or (a, b) -> (size_of_pred a) + (size_of_pred b) + 1
  
(* Pretty printer for the debugger. *)
let pp_pred ch p = Format.fprintf ch "%s" (print_pred p)

type pol =
    | Pred of pred
    | Modify of Field.t * int
    | Table of TableName.t * pol_t
    (* if a then p else q == a; p + -a; q. *)
    | IfThen of pred * pol * pol
    | Union of pol * pol
    | Seq of pol * pol
    | Con of pol * FieldSet.t * FieldSet.t * pol

let rec print_pol (p : pol) : string =
  match p with
  | Pred pr -> print_pred pr
  | Modify (f, v) -> Printf.sprintf "%s <- %d" (print_field f) v
  | Table (n, t) -> (TableName.print n) ^ " : " ^ (print_pol_t t)
  | IfThen (a, p_true, p_false) ->
    Printf.sprintf "if %s then %s else %s"
      (print_pred a)
      (print_pol p_true)
      (print_pol p_false)
  | Union (p1, p2) ->
    let rv1 = print_pol p1 in
    let rv2 = print_pol p2 in
    rv1 ^ " + " ^ rv2
  | Seq (p1, p2) ->
    let rv1 = print_pol p1 in
    let rv2 = print_pol p2 in
    rv1 ^ "; " ^ rv2
  | Con (p1, w1, w2, p2) ->
    let p1_s = print_pol p1 in
    let p2_s = print_pol p2 in
    let w1_s = print_fields w1 in
    let w2_s = print_fields w2 in
    Printf.sprintf "%s %s_||_%s %s" p1_s w1_s w2_s p2_s

let rec size_of_pol (pol : pol) : int =
  match pol with
  | Pred a -> size_of_pred a
  | Modify _ | Table _ -> 1
  | IfThen (a, p, q) ->
    (size_of_pred a) + (size_of_pol p) + (size_of_pol q) + 1
  | Union (p, q) | Seq (p, q) | Con (p, _, _, q) ->
    (size_of_pol p) + (size_of_pol q) + 1

(* Pretty printer for the debugger. *)
let pp_pol ch p = Format.fprintf ch "%s" (print_pol p)

(* Return a set of the tables and their types in the policy.*)
let rec tables_of_pol (p : pol) : TableSet.t =
  match p with
  | Pred _
  | Modify _ -> TableSet.empty
  | Table (x, _) -> TableSet.singleton x
  | IfThen (_, p1, p2)
  | Union (p1, p2)
  | Seq (p1, p2)
  | Con (p1, _, _, p2) -> TableSet.union (tables_of_pol p1) (tables_of_pol p2)

(* Build an if statement. *)
let mkIf (a : pred) (p_true : pol) (p_false : pol) : pol =
  if p_true = p_false then
    p_true
  else
    match a with
    | Id -> p_true
    | Drop -> p_false
    | _ -> IfThen (a, p_true, p_false)

(* Simple optimizing constructors: PREDICATES. *)
let (<^>) (a : pred) (b : pred) : pred =
  match a,b with
  | Id, _ -> b
  | _, Id -> a
  | Drop, _ -> Drop
  | _, Drop -> Drop
  | _, Neg b' -> if a = b' then Drop else And (a,b)
  | Neg a', _ -> if b = a' then Drop else And (a,b)
  | _, _ -> And (a,b)

let (<||>) (a : pred) (b : pred) : pred =
  match a,b with
  | Id, _ -> Id
  | _, Id -> Id
  | Drop, _ -> b
  | _, Drop -> a
  | _, Neg b' -> if a = b' then Id else Or (a,b)
  | Neg a', _ -> if b = a' then Id else Or (a,b)
  | _, _ -> Or (a,b)

let (<!>) (a : pred) : pred =
  match a with
  | Id -> Drop
  | Drop -> Id
  | Neg b -> b
  | _ -> Neg a

(* Simple optimizing constructors: POLICIES. *)
let (<.>) (p : pol) (q : pol) : pol =
  match p,q with
  | Pred Id, _ -> q
  | _, Pred Id -> p
  | Pred Drop, _ -> Pred Drop
  | _, Pred Drop -> Pred Drop
  | _, _ -> Seq (p,q)

let (<+>) (p : pol) (q : pol) : pol =
  match p,q with
  | Pred Drop, _ -> q
  | _, Pred Drop -> p
  | Seq (Pred a, p'), Seq (Pred (Neg b), q') ->
    if a = b then mkIf a p' q' else Union (p,q)
  | Seq (Pred (Neg a), p'), Seq (Pred b, q') ->
    if a = b then mkIf b q' p' else Union (p,q)
  | _, _ -> Union (p,q)



(* Replace pred combinators pol combinators. *)
let rec pred_lift (pr : pred) : pol =
  match pr with
  | Id
  | Drop
  | Match _
  | Neg _ -> Pred pr
  | And (a, b) -> (pred_lift a) <.> (pred_lift b)
  | Or (a,b) -> (pred_lift a) <+> (pred_lift b)


let rec pol_contains_table (p : pol) : bool =
  match p with
  | Pred _
  | Modify _ -> false
  | Table _ -> true
  | IfThen (_,p1,p2)
  | Union (p1,p2)
  | Seq (p1,p2)
  | Con (p1, _, _, p2) ->
    pol_contains_table p1 || pol_contains_table p2

let desugar_if ((a,p,q) : pred * pol * pol) : pol =
  let left = Seq (Pred a, p) in
  let right = Seq (Pred (Neg a), q) in
  Union (left, right)

let rec check_pred (pr : pred) (fs : FieldSet.t) : bool =
  match pr with
  | Id
  | Drop -> true
  | Match (f,_) -> FieldSet.mem f fs
  | Neg pr' -> check_pred pr' fs
  | And (pr1, pr2)
  | Or (pr1, pr2) ->
    let rv1 = check_pred pr1 fs in
    let rv2 = check_pred pr2 fs in
    rv1 && rv2

let rec check_pol (p : pol) (p_t : pol_t) : 
  (bool, string) Result.t =
  let (rs, ws) = p_t in
  (* Check for well-formedness of the type. *)
  if not (inter_is_empty rs ws) then
    Error
      (Printf.sprintf
        "malformed type: (%s, %s)"
        (print_fields rs)
        (print_fields ws))
  else

  (* Typecheck_pol the term. *)
  match p with
  | Pred pred -> 
    let fields = FieldSet.union rs ws in
    Ok (check_pred pred fields)
  | Modify (f,_) -> Ok (FieldSet.mem f ws)
  | Table (_, (table_rs, table_ws)) ->
    let rv1 = FieldSet.subset table_rs rs in
    let rv2 = FieldSet.subset table_ws ws in
    Ok (rv1 && rv2)
  | IfThen (a,p',q') ->
    let a_res = check_pred a (FieldSet.union rs ws) in
    check_pol p' p_t >>= fun rv1 ->
    check_pol q' p_t >>= fun rv2 ->
    Ok (a_res && rv1 && rv2)
  | Union (p1, p2)
  | Seq (p1, p2) ->
    check_pol p1 (rs, ws) >>= fun rv1 ->
    check_pol p2 (rs, ws) >>= fun rv2 ->
    Ok (rv1 && rv2)
  | Con (p1, ws1, ws2, p2) ->
    (* Pick the most permissive read sets. *)
    let rs1 = FieldSet.diff rs ws1 in
    let rs2 = FieldSet.diff rs ws2 in
    check_pol p1 (rs1, ws1) >>= fun rv1 ->
    check_pol p2 (rs2, ws2) >>= fun rv2 ->
    Ok (    rv1
         && rv2
         && (inter_is_empty ws1 ws2)
         && (inter_is_empty ws1 rs2)
         && (inter_is_empty ws2 rs1) )

let rec infer_pred (pr : pred) : FieldSet.t =
  match pr with
  | Id
  | Drop -> FieldSet.empty
  | Match (f,_) -> FieldSet.singleton f
  | Neg pr' -> infer_pred pr'
  | And (pr1, pr2)
  | Or (pr1, pr2) ->
    let rv1 = infer_pred pr1 in
    let rv2 = infer_pred pr2 in
    FieldSet.union rv1 rv2

let rec infer_pol (p : pol) : pol_t =
  match p with
  | Pred pred -> (infer_pred pred, FieldSet.empty)
  | Modify (f,_) -> (FieldSet.empty, FieldSet.singleton f)
  | Table (_, table_t) -> table_t
  | IfThen (a,p1,p2) ->
    let rsa = infer_pred a in
    let (rs1, ws1) = infer_pol p1 in
    let (rs2, ws2) = infer_pol p2 in
    let ws = FieldSet.union ws1 ws2 in
    let rs = FieldSet.diff (FieldSet.union rsa (FieldSet.union rs1 rs2)) ws in
    (rs, ws)
  | Union (p1, p2)
  | Seq (p1, p2) ->
    let (rs1, ws1) = infer_pol p1 in
    let (rs2, ws2) = infer_pol p2 in
    let ws = FieldSet.union ws1 ws2 in
    let rs = FieldSet.diff (FieldSet.union rs1 rs2) ws in
    (rs, ws)
  | Con (p1, w1, w2, p2) ->
    let (rs1, ws1) = infer_pol p1 in
    let (rs2, ws2) = infer_pol p2 in
    let ws = FieldSet.union ws1 ws2 in
    let rs = FieldSet.diff (FieldSet.union rs1 rs2) ws in
    (* Check for failure. *)
    let msg = Printf.sprintf "cannot be typed: (%s)" (print_pol p) in 
    if not (FieldSet.subset ws1 w1) || not (FieldSet.subset ws2 w2) then
      raise (Failure msg)
    else if not (FieldSet.inter rs1 ws2 |> FieldSet.is_empty)
         || not (FieldSet.inter rs2 ws1 |> FieldSet.is_empty)
         || not (FieldSet.inter w1 w2 |> FieldSet.is_empty) then
      raise (Failure msg)
    else
      (rs, ws)

let rec extract_concurrency (p : pol) : pol =
  (* TODO: make this more efficient by not recomputing types for every
   * subtree. *)
  let is_disjoint s1 s2 = FieldSet.inter s1 s2 |> FieldSet.is_empty in
  match p with
  | Pred _
  | Modify _
  | Table _
  | IfThen _
  | Union _ -> p
  | Seq (p1,p2) ->
    let p1' = extract_concurrency p1 in
    let p2' = extract_concurrency p2 in
    let r1, w1 = infer_pol p1' in
    let r2, w2 = infer_pol p2' in
    if (    is_disjoint r1 w2
         && is_disjoint r2 w1
         && is_disjoint w1 w2 ) then
      Con (p1', w1, w2, p2')
    else
      Seq (p1', p2')
  | Con (p1, w1, w2, p2) ->
    let p1' = extract_concurrency p1 in
    let p2' = extract_concurrency p2 in
    Con (p1', w1, w2, p2')
