(*
 * The MIT License (MIT)
 * 
 * Copyright (c) 2015 Cole Schlesinger
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *)


open Syntax
open Semantics
open Compiler
open Equivalence
open Test_utils
open My_core


(* Set up translation validation: evaluate the input and output and check that
 * they produce equivalent results. *)
let mkMCTV
  (p : pol)
  (s : ((TableName.t * pol_t) -> FieldSet.t))
  (b : table_binding)
  (pks : PacketSet.t) : test_t =
  let verbose = false in
  fun () -> 
  let metadata = MetadataField.gen_many (estimate_mc_metadata p) in
  let m, n, theta = pipeline s metadata p mod_stage_id in

  (* TODO: remove *)
  let () =
    if verbose then
      let () = Printf.printf "\n" in
      let () = Printf.printf "SOURCE: (%s)\n" (print_pol p) in
      let () = Printf.printf "PRODUCED: (%s); (%s)\n"
        (print_mod_stage m) (print_egress_stage n) in
      ()
    else
      ()
    in

  let sub_p = sub_pol b p in
  let sub_m = sub_mods (theta b) m in
  let sub_n = sub_egress (theta b) n in
  let sub_p_out =
    (mod_stage_to_pol sub_m) <.> (egress_stage_to_pol sub_n) in

  (* TODO: remove *)
  let () =
    if verbose then
      let () = Printf.printf "SOURCE SUBBED: (%s)\n" (print_pol sub_p) in
      let () = Printf.printf "SUBBED: (%s)\n"
        (print_pol sub_p_out) in
      ()
    else
      ()
    in

  let zeroed_pks = zero_metadata pks metadata in
  eval sub_p pks >>= fun pks_source ->
  eval sub_p_out zeroed_pks >>= fun pks_target ->
  let zeroed_pks_source = zero_metadata pks_source metadata in
  let zeroed_pks_target = zero_metadata pks_target metadata in
  if PacketSet.equal zeroed_pks_source zeroed_pks_target then
    Ok None
  else
    let msg =
      Printf.sprintf "in: (%s), source out: (%s), target out: (%s)"
        (print_packets pks)
        (print_packets zeroed_pks_source)
        (print_packets zeroed_pks_target) in
    Ok (Some msg)

(* Set up translation validation: evaluate the input and output and check that
 * they produce equivalent results. *)
let mkFETV
  (f : Field.t)
  (p : pol)
  (s : ((TableName.t * pol_t) -> FieldSet.t))
  (b : table_binding)
  (pks : PacketSet.t) : test_t =
  let verbose = false in
  fun () -> 
  let metadata = MetadataField.gen_many (estimate_fe_metadata p) in
  let m, p', theta = ext f s metadata p mod_stage_id in

  (* TODO: remove *)
  let () =
    if verbose then
      let () = Printf.printf "\n" in
      let () = Printf.printf "SOURCE: (%s)\n" (print_pol p) in
      let () = Printf.printf "PRODUCED: (%s); (%s)\n"
        (print_pol (mod_stage_to_pol m)) (print_pol p') in
      ()
    else
      ()
    in

  let sub_p = sub_pol b p in
  let sub_m = sub_mods (theta b) m in
  let sub_p' = sub_pol (theta b) p' in
  let sub_p_out = (mod_stage_to_pol sub_m) <.> sub_p' in

  (* TODO: remove *)
  let () =
    if verbose then
      let () = Printf.printf "SOURCE SUBBED: (%s)\n" (print_pol sub_p) in
      let () = Printf.printf "SUBBED: (%s)\n"
        (print_pol sub_p_out) in
      ()
    else
      ()
    in

  let zeroed_pks = zero_metadata pks metadata in
  eval sub_p pks >>= fun pks_source ->
  eval sub_p_out zeroed_pks >>= fun pks_target ->
  let zeroed_pks_source = zero_metadata pks_source metadata in
  let zeroed_pks_target = zero_metadata pks_target metadata in
  if PacketSet.equal zeroed_pks_source zeroed_pks_target then
    Ok None
  else
    let msg =
      Printf.sprintf "in: (%s), source out: (%s), target out: (%s)"
        (print_packets pks)
        (print_packets zeroed_pks_source)
        (print_packets zeroed_pks_target) in
    Ok (Some msg)

(* Build a table binding from a list. *)
let mkTableBinding
  (tables : (TableName.t * pol) list) : table_binding =
  List.fold_left (fun map (x, p) ->
    let size_p = height p TableBinding.empty in
    TableBinding.add x (Policy p, size_p) map)
    TableBinding.empty
    tables

(* Makes a table binding with estimates and the policy Id. *)
let mkTableEstimate
  (tables : ((TableName.t * pol_t) * int) list) : table_binding =
  List.fold_left (fun map ((x, _), n) ->
    TableBinding.add x (Policy (Pred Id), n) map)
    TableBinding.empty
    tables

let mkTest 
  (pks : PacketSet.t)
  (p : pol)
  (tables : (TableName.t * pol) list) : test_t =
  fun () ->

  (* Prepare substitutions. *)
  let b = mkTableBinding tables in
  let sub_p = sub_pol b p in

  let verbose = false in
  let () =
    if verbose then
      let () =
        Printf.printf
          "========================================================\n" in
      let () = Printf.printf "SOURCE: (%s)\n" (print_pol p) in
      let () = Printf.printf "SOURCE SUBBED: (%s)\n" (print_pol sub_p) in
      let () = Printf.printf "\n" in
      ()
    else
      ()
    in

  (* Multicast consolidation. *)
  let mc_smap =
    List.fold_left
      (fun map (x, px) ->
        let metadata =
          MetadataField.gen_many (estimate_mc_metadata px) in
        TableBinding.add x metadata map)
      TableBinding.empty
      tables in
  let mc_s =
    (fun (x,t) ->
      try
        TableBinding.get x mc_smap
      with
        Not_found ->
          let msg = Printf.sprintf "mc_s: not found: (%s, %s)"
            (TableName.print x) (print_pol_t t) in
          raise (Failure msg)) in
  let mc_metadata =
    MetadataField.gen_many (estimate_mc_metadata p) in
  let m1, n1, theta1 = pipeline mc_s mc_metadata p mod_stage_id in
  let n1_pol = egress_stage_to_pol n1 in
  let n1_tables = tables_of_pol n1_pol in

  let () =
    if verbose then
      let () = Printf.printf "\n" in
      let () = Printf.printf "CONSOLIDATION: (%s)\n" (print_pol (mod_stage_to_pol m1)) in
      let () = Printf.printf "CONSOLIDATION SUBBED: (%s)\n\n"
        (print_pol (mod_stage_to_pol (sub_mods (theta1 b) m1))) in
      let () = Printf.printf "EGRESS: (%s)\n" (print_pol n1_pol) in
      let () = Printf.printf "EGRESS SUBBED: (%s)\n"
        (print_pol (sub_pol_forced (theta1 b) n1_pol)) in
      let () = Printf.printf "\n" in
      ()
    else
      ()
    in

  (* Validate MC. *)
  let mc_s_metadata =
    TableBinding.fold
      (fun _ fs acc -> FieldSet.union fs acc)
      mc_smap
      FieldSet.empty in
  let all_metadata =
    FieldSet.union mc_s_metadata mc_metadata |>
    FieldSet.union MetadataField.special_metadata in
  let sub_mc = (sub_pol_forced (theta1 b) (mod_stage_to_pol m1)) <.> (sub_pol_forced (theta1 b) n1_pol) in

  let () = match setupTestEval pks all_metadata sub_p sub_mc with
  | Error msg -> raise (Failure ("(MC) " ^ msg))
  | Ok (Some msg) -> raise (Failure ("(MC) " ^ msg))
  | Ok None -> ()
  in

  (* Field extraction. *)
  let b_after_mc = theta1 b in
  let (fe_smap : FieldSet.t TableBinding.t) =
    TableBinding.mapi
      (fun k (res, _) ->
        let pol = match res with
          | Policy p -> p
          | ModStage m -> mod_stage_to_pol m
          | EgressStage n -> egress_stage_to_pol n
          in
        if TableSet.mem k n1_tables then
          MetadataField.gen_many (estimate_fe_metadata pol)
        else
          FieldSet.empty)
      b_after_mc in
  let fe_s = (fun (x, _) -> TableBinding.get x fe_smap) in
  let fe_metadata =
    MetadataField.gen_many (estimate_fe_metadata n1_pol) in
  let e1, n2, theta2 =
    ext "OutPort" fe_s fe_metadata n1_pol mod_stage_id in

  let () =
    if verbose then
      let () = Printf.printf "\n" in
      let () = Printf.printf "EXTRACTION: (%s)\n" (print_pol (mod_stage_to_pol e1)) in
      let () = Printf.printf "EXTRACTION SUBBED: (%s)\n\n"
        (print_pol (mod_stage_to_pol (sub_mods (theta1 b |> theta2) e1))) in
      let () = Printf.printf "EXTRACTED EGRESS: (%s)\n" (print_pol n2) in
      let () = Printf.printf "EXTRACTED EGRESS SUBBED: (%s)\n"
        (print_pol (sub_pol_forced (theta1 b |> theta2) n2)) in
      let () = Printf.printf "\n" in
      ()
    else
      ()
    in

  (* Validate FE. *)
  let fe_s_metadata =
    TableBinding.fold
      (fun _ fs acc -> FieldSet.union fs acc)
      fe_smap
      FieldSet.empty in
  let all_metadata =
    FieldSet.union mc_s_metadata mc_metadata |>
    FieldSet.union fe_s_metadata |>
    FieldSet.union fe_metadata |>
    FieldSet.union MetadataField.special_metadata in
  let fe_theta = (fun b -> (theta1 b) |> theta2) in
  let sub_fe = (sub_pol_forced (fe_theta b) (mod_stage_to_pol m1))
    <.> (sub_pol_forced (fe_theta b) (mod_stage_to_pol e1))
    <.> (sub_pol_forced (fe_theta b) n2) in

  let () =
    let emsg = Printf.sprintf "(FE): (%s)\n" (print_pol sub_fe) in
    match setupTestEval pks all_metadata sub_p sub_fe with
    | Error msg -> raise (Failure (emsg ^ msg))
    | Ok (Some msg) -> raise (Failure (emsg ^ msg))
    | Ok None -> ()
    in


  (* Combine consolidation and extraction stages. *)
  let s1, s2, s3, theta3 = combine_mc_and_fe m1 e1 in

  (* Get generated table size estimates. *)
  let _, theta_tabled = insert_tables n2 in
  let theta = (fun b -> (theta1 b) |> theta2 |> theta3 |> theta_tabled) in

  (* For table fitting.
  let b_out = theta b in
  let partitions = partition (extract_concurrency n2_tabled) b_out in

  let pol_of_partitions =
    let flattened = (List.map snd partitions) |> List.flatten in
    List.fold_left 
      (fun acc (x,t) -> Table (x,t) <.> acc)
      (Pred Id)
      (List.rev flattened) in
  let combined = s1 <.> s2 <.> s3 in
  let p_out = combined <.> pol_of_partitions in
  *)
  let combined = s1 <.> s2 <.> s3 in
  let p_out = combined <.> n2 in

  (* Do substitution. *)
  let sub_p_out = sub_pol_forced (theta b) p_out in

  (* Poor/lazy man's debugging. *)
  let () =
    if verbose then
      let () =
        Printf.printf "TARGET: (%s);(%s);(%s);(%s)\n"
          (print_pol s1)
          (print_pol s2)
          (print_pol s3)
          (print_pol n2)
          in
      let () =
        Printf.printf "TARGET SUBBED: (%s)\n" (print_pol sub_p_out) in
      ()
    else
      ()
    in

  (* Collect all the metadata fields that might have been used. *)
  let all_metadata =
    FieldSet.union mc_s_metadata fe_s_metadata |>
    FieldSet.union mc_metadata |>
    FieldSet.union fe_metadata |>
    FieldSet.union MetadataField.special_metadata in

  (* Run p and p_out on pks and test the results. *)
  setupTestEval pks all_metadata sub_p sub_p_out

let mkTests
  (pks : PacketSet.t)
  (tests : (string * pol * ((TableName.t * pol) list)) list)
  : (string * test_t) list =
  List.map
    (fun (name, p, tables) -> (name, mkTest pks p tables))
    tests

let test_specialize_1 () : test_result =
  let f = "SrcMAC" in
  let v = 1 in
  let a = And (Match ("SrcMAC", 1), Neg (Match ("DstMAC", 1))) in
  let b = specialize f v a in
  let b_expected = Neg (Match ("DstMAC", 1)) in
  if b = b_expected then
    Ok None
  else
    let msg =
      Printf.sprintf
        "expected: (%s),  got (%s)"
        (print_pred b_expected)
        (print_pred b) in
    Ok (Some msg)

let spec_tests =
  [ ("specialize 1", test_specialize_1)
  ]


(*** MULTICAST CONSOLIDATION TESTS ****************************************)

let mkMCTest
  (p_in : pol)
  (m_in : mod_stage)
  (e_expected : egress_stage)
  (m_expected : mod_stage) : test_t =
  fun () ->
  let m_out, e_out, _ = pipeline (fun _ -> FieldSet.empty) FieldSet.empty p_in m_in in
  if (m_out = m_expected) && (e_out = e_expected) then
    Ok None
  else
    let msg =
      Printf.sprintf "expected: (%s | %s), got: (%s | %s)"
        (print_egress_stage e_expected) (print_mod_stage m_expected)
        (print_egress_stage e_out) (print_mod_stage m_out) in
    Ok (Some msg)

let mctest_id =
  let p = Pred Id in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  mkMCTV p s b pks

let mctest_drop =
  let p = Pred Drop in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  mkMCTV p s b pks

let mctest_1 =
  let p_in = Pred (Match ("DstMAC", 1)) in
  let m_in = mod_stage_id in
  let p_out = egress_pol p_in in
  let (nf : NF.t) = nf_of_pred (Match ("DstMAC", 1)) in
  let m_out = [(nf, Mods [])] in
  mkMCTest p_in m_in p_out m_out

let mctest_2 =
  let p_in = Modify ("SrcMAC", 1) in
  let m_in = mod_stage_id in
  let p_out = egress_pol p_in in
  let m_out = m_in in
  mkMCTest p_in m_in p_out m_out

let mctest_3 =
  let p = Pred (Match ("DstMAC", 1)) in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  mkMCTV p s b pks

let mctest_4 =
  let p = Pred (Match ("SrcMAC", 1)) in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  mkMCTV p s b pks

let mctest_5 : test_t =
  let p' = Pred (Match ("DstMAC", 1)) in
  let tp' = infer_pol p' in
  let x = TableName.gen () in
  let p = Table (x, tp') in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, p')] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  mkMCTV p s b pks

let mctest_6 : test_t =
  let p' = Pred (Match ("SrcMAC", 1)) in
  let tp' = infer_pol p' in
  let x = TableName.gen () in
  let p = Table (x, tp') in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, p')] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  mkMCTV p s b pks

let mctest_7 : test_t =
  let p' = Modify ("SrcMAC", 1) in
  let tp' = infer_pol p' in
  let x = TableName.gen () in
  let p = Table (x, tp') in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, p')] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  mkMCTV p s b pks

let mctest_8 : (string * test_t) =
  let p' = Modify ("SrcMAC", 1) in
  let tp' = infer_pol p' in
  let x = TableName.gen () in
  let p = IfThen(Match ("DstMAC", 1), Table (x, tp'), Pred Drop) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, p')] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("MC - if - 1", mkMCTV p s b pks)

let mctest_9 : (string * test_t) =
  let p' = Modify ("SrcMAC", 1) in
  let tp' = infer_pol p' in
  let x = TableName.gen () in
  let p = Union (Table (x, tp'), Pred Drop) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, p')] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("MC - union - 1", mkMCTV p s b pks)

let mctest_10 : (string * test_t) =
  let p' = Modify ("SrcMAC", 1) in
  let tp' = infer_pol p' in
  let x = TableName.gen () in
  let p = Union (Table (x, tp'), Pred Id) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, p')] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("MC - union - 2", mkMCTV p s b pks)

let mctest_11 : (string * test_t) =
  let p' = Modify ("SrcMAC", 1) in
  let tp' = infer_pol p' in
  let x = TableName.gen () in
  let p = Seq (Table (x, tp'), Pred Id) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, p')] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("MC - seq - 1", mkMCTV p s b pks)

let mctest_12 : (string * test_t) =
  let p' = Modify ("SrcMAC", 1) in
  let tp' = infer_pol p' in
  let x = TableName.gen () in
  let p = Seq (Table (x, tp'), Pred Drop) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, p')] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("MC - seq - 2", mkMCTV p s b pks)

let mctest_13 : (string * test_t) =
  let p' = Modify ("SrcMAC", 1) in
  let tp' = infer_pol p' in
  let x = TableName.gen () in
  let p = Pred (Match ("SrcMAC", 2)) <.> (Table (x, tp') <+> Pred Id) <.> Pred Id in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, p')] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("MC - seq + union - 1", mkMCTV p s b pks)


(*** FIELD EXTRACTION TESTS ***********************************************)

let fetest_id : (string * test_t) =
  let f = "SrcMAC" in
  let p = Pred Id in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - id - 1", mkFETV f p s b pks)

let fetest_drop : (string * test_t) =
  let f = "SrcMAC" in
  let p = Pred Drop in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - drop - 1", mkFETV f p s b pks)

let fetest_1 : (string * test_t) =
  let f = "SrcMAC" in
  let p = Pred (Match ("SrcMAC", 1) <||> Match ("DstMAC", 2)) in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - pred - 1", mkFETV f p s b pks)

let fetest_2 : (string * test_t) =
  let f = "SrcMAC" in
  let p = Modify ("SrcMAC", 3) in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - mod - 1", mkFETV f p s b pks)

let fetest_3 : (string * test_t) =
  let f = "SrcMAC" in
  let p = Modify ("DstMAC", 3) in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - mod - 2", mkFETV f p s b pks)

let fetest_3a : (string * test_t) =
  let f = "OutPort" in
  let p = Modify ("OutPort", 3) in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks =
    mkPacketSet
      [mkPacket [("SrcMAC", 2); ("DstMAC", 1); ("OutPort", 0)]] in
  ("FE - mod - 3", mkFETV f p s b pks)

let fetest_4 : (string * test_t) =
  let f = "SrcMAC" in
  let px = Pred Id in
  let x = TableName.gen () in
  let tx = infer_pol px in
  let p = Table (x, tx) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, px)] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - table id - 1", mkFETV f p s b pks)

let fetest_5 : (string * test_t) =
  let f = "SrcMAC" in
  let px = Pred Drop in
  let tx = infer_pol px in
  let x = TableName.gen () in
  let p = Table (x, tx) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, px)] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - table drop - 1", mkFETV f p s b pks)

let fetest_6 : (string * test_t) =
  let f = "SrcMAC" in
  let px = Pred (Match ("SrcMAC", 1) <||> Match ("DstMAC", 2)) in
  let tx = infer_pol px in
  let x = TableName.gen () in
  let p = Table (x, tx) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, px)] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - table pred - 1", mkFETV f p s b pks)

let fetest_7 : (string * test_t) =
  let f = "SrcMAC" in
  let px = Modify ("SrcMAC", 3) in
  let x = TableName.gen () in
  let tx = infer_pol px in
  let p = Table (x, tx) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, px)] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - table mod - 1", mkFETV f p s b pks)

let fetest_8 : (string * test_t) =
  let f = "SrcMAC" in
  let px = Modify ("DstMAC", 3) in
  let x = TableName.gen () in
  let tx = infer_pol px in
  let p = Table (x, tx) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, px)] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - table mod - 2", mkFETV f p s b pks)

let fetest_9 : (string * test_t) =
  let f = "DstMAC" in
  let p = IfThen(Match ("DstMAC", 1), Modify ("DstMAC", 3), Pred Id) in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - if - 1", mkFETV f p s b pks)

let fetest_10 : (string * test_t) =
  let f = "SrcMAC" in
  let p = IfThen(Match ("SrcMAC", 2), Modify ("DstMAC", 3), Pred Id) in
  let s = (fun _ -> FieldSet.empty) in
  let b = TableBinding.empty in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - if - 1", mkFETV f p s b pks)


let fetest_11 : (string * test_t) =
  let f = "DstMAC" in
  let px = Modify ("DstMAC", 3) in
  let x = TableName.gen () in
  let tx = infer_pol px in
  let p = IfThen(Match ("DstMAC", 1), Table (x, tx), Pred Id) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, px)] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - table if - 1", mkFETV f p s b pks)

let fetest_12 : (string * test_t) =
  let f = "SrcMAC" in
  let px = Modify ("DstMAC", 3) in
  let x = TableName.gen () in
  let tx = infer_pol px in
  let p = IfThen(Match ("SrcMAC", 2), Table (x, tx), Pred Id) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, px)] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - table if - 2", mkFETV f p s b pks)

let fetest_13 : (string * test_t) =
  let f = "SrcMAC" in
  let px = Modify ("DstMAC", 3) in
  let x = TableName.gen () in
  let tx = infer_pol px in
  let p =
    Seq (Pred (Match ("DstMAC", 1)),
         IfThen(Match ("SrcMAC", 2), Table (x, tx), Pred Id)) in
  let s = (fun _ -> FieldSet.empty) in
  let b = mkTableBinding [(x, px)] in
  let pks = mkPacketSet [mkPacket [("SrcMAC", 2); ("DstMAC", 1)]] in
  ("FE - table seq if - 1", mkFETV f p s b pks)


(*** TABLE FITTING TESTS **************************************************)

let print_s (s : int option array array) : string =
  print_matrix 
    (fun opt ->
      match opt with
      | Some n -> string_of_int n
      | None -> "None" )
    s

let make_build_cost_tables_test xs estimates m_expected s_expected : test_t =
  fun () ->
  let m_out, s_out = build_cost_tables xs estimates in
  if (CostArray.equal m_out m_expected)
      && (matrix_equal (=) s_out s_expected) then
    Ok None
  else
    let msg = Printf.sprintf "expected: \n(%s), \n(%s), got: \n(%s), \n(%s)"
      (CostArray.print m_expected)
      (print_s s_expected)
      (CostArray.print m_out)
      (print_s s_out)
      in
    Ok (Some msg)

let teblocks_1 : test_t =
  fun () ->
  let t = (FieldSet.singleton "OutPort", FieldSet.empty) in
  let x = TableName.gen () in
  let xs = [x, t] in
  let table_estimates = mkTableEstimate [(List.hd xs, 1)] in
  let blocks_res =
    (blocks (Table (x,t)) table_estimates) in
  if blocks_res = 1 then
    Ok None
  else
    let msg =
      Printf.sprintf "expected: 1, got: (%d)" blocks_res in
    Ok (Some msg)

let tetest_single : test_t =
  fun () ->
  let t = (FieldSet.singleton "OutPort", FieldSet.empty) in
  let x = TableName.gen () in
  let xs = [x, t] in
  let table_estimates = mkTableEstimate [(List.hd xs, 1)] in
  let m, _ = build_cost_tables xs table_estimates in
  if (CostArray.to_list m) = [[Cost 1]] then
    Ok None
  else
    let msg =
      Printf.sprintf "expected: 1, got: (%s)" (CostArray.print m) in
    Ok (Some msg)

let tetest_1 : test_t =
  let x1 = TableName.gen () in
  let x2 = TableName.gen () in
  let t = (FieldSet.singleton "OutPort", FieldSet.empty) in
  let xs = [(x1, t); (x2, t)] in
  let estimates =
    mkTableEstimate [((x1, t), 1); ((x2, t), 1)] in
  let n = List.length xs in
  let m_expected = CostArray.make n in
  let s_expected = Array.make_matrix n n None in
  let () = CostArray.set (0,0) (Cost 1) m_expected in
  let () = CostArray.set (1,1) (Cost 1) m_expected in
  let () = CostArray.set (0,1) (Cost 1) m_expected in
  make_build_cost_tables_test xs estimates m_expected s_expected

let tetest_2 : test_t =
  let x1 = TableName.gen () in
  let x2 = TableName.gen () in
  let t = (FieldSet.singleton "OutPort", FieldSet.empty) in
  let xs = [(x1, t); (x2, t)] in
  let estimates =
    mkTableEstimate [((x1, t), 1); ((x2, t), 1)] in
  let n = List.length xs in
  let m_expected = CostArray.make n in
  let s_expected = Array.make_matrix n n None in
  let () = CostArray.set (0,0) (Cost 1) m_expected in
  let () = CostArray.set (1,1) (Cost 1) m_expected in
  let () = CostArray.set (0,1) (Cost 1) m_expected in
  make_build_cost_tables_test xs estimates m_expected s_expected

let tetest_3 : test_t =
  let x1 = TableName.gen () in
  let x2 = TableName.gen () in
  let t1 = (FieldSet.singleton "SrcIP", FieldSet.empty) in
  let t2 = (FieldSet.singleton "DstIP", FieldSet.empty) in
  let xs = [(x1, t1); (x2, t2)] in
  let estimates =
    mkTableEstimate [((x1, t1), 2048); ((x2, t2), 2048)] in
  let n = List.length xs in
  let m_expected = CostArray.make n in
  let s_expected = Array.make_matrix n n None in
  let () = CostArray.set (0,0) (Cost 1) m_expected in
  let () = CostArray.set (1,1) (Cost 1) m_expected in
  let () = CostArray.set (0,1) (Cost 2) m_expected in
  let () = s_expected.(0).(1) <- Some 0 in
  make_build_cost_tables_test xs estimates m_expected s_expected

let tetest_4 : test_t =
  let x1 = TableName.gen () in
  let x2 = TableName.gen () in
  let x3 = TableName.gen () in
  let t1 = (FieldSet.singleton "SrcIP", FieldSet.empty) in
  let t2 = (FieldSet.singleton "DstIP", FieldSet.empty) in
  let t3 = (FieldSet.empty, FieldSet.singleton "OutPort") in
  let xs = [(x1, t1); (x2, t2); (x3, t3)] in
  let estimates =
    mkTableEstimate [((x1, t1), 2048); ((x2, t2), 1024); ((x3, t3), 1024)] in
  let n = List.length xs in
  let m_expected = CostArray.make n in
  let s_expected = Array.make_matrix n n None in
  let () = CostArray.set (0,0) (Cost 1) m_expected in
  let () = CostArray.set (1,1) (Cost 1) m_expected in
  let () = CostArray.set (2,2) (Cost 1) m_expected in
  let () = CostArray.set (0,1) (Cost 2) m_expected in
  let () = CostArray.set (1,2) (Cost 2) m_expected in
  let () = CostArray.set (0,2) (Cost 3) m_expected in
  let () = s_expected.(0).(1) <- Some 0 in
  let () = s_expected.(1).(2) <- Some 0 in
  let () = s_expected.(0).(2) <- Some 0 in
  make_build_cost_tables_test xs estimates m_expected s_expected

let tetest_5 : test_t =
  let x1 = TableName.gen () in
  let x2 = TableName.gen () in
  let x3 = TableName.gen () in
  let t1 = (FieldSet.singleton "SrcIP", FieldSet.empty) in
  let t2 = (FieldSet.singleton "DstIP", FieldSet.empty) in
  let t3 = (FieldSet.singleton "InPort", FieldSet.empty) in
  let xs = [(x1, t1); (x2, t2); (x3, t3)] in
  let estimates =
    mkTableEstimate [((x1, t1), 2048); ((x2, t2), 1024); ((x3, t3), 32)] in
  let n = List.length xs in
  let m_expected = CostArray.make n in
  let s_expected = Array.make_matrix n n None in
  let () = CostArray.set (0,0) (Cost 1) m_expected in
  let () = CostArray.set (1,1) (Cost 1) m_expected in
  let () = CostArray.set (2,2) (Cost 1) m_expected in
  let () = CostArray.set (0,1) (Cost 2) m_expected in
  let () = CostArray.set (1,2) (Cost 1) m_expected in
  let () = CostArray.set (0,2) (Cost 2) m_expected in
  let () = s_expected.(0).(1) <- Some 0 in
  let () = s_expected.(1).(2) <- None in
  let () = s_expected.(0).(2) <- Some 0 in
  make_build_cost_tables_test xs estimates m_expected s_expected


(*** COMBINED TESTS *******************************************************)

let test_pks = mkPacketSet
  [ mkPacket [ ("SrcMAC", 2); ("DstMAC", 1); ("OutPort", 0) ]
  ; mkPacket [ ("SrcMAC", 1); ("DstMAC", 2); ("OutPort", 0) ]
  ]
let non_table_test_pols =
  [ ("combined 1", Modify ("OutPort", 1), [])
  ; ("combined 2"
    , mkIf (Match ("DstMAC", 1)) (Modify ("OutPort", 1)) (Pred Drop)
    , [])
  ; ("combined 3"
    , Pred (Match ("DstMAC", 2)) <.> Modify ("OutPort", 2)
    , [])
  ; ("combined 4"
    , Modify ("OutPort", 1) <.> Modify ("OutPort", 2)
    , [])
  ; ("combined 5"
    , (mkIf (Match ("DstMAC", 1)) (Modify ("OutPort", 1)) (Pred Drop))
      <.> (mkIf (Match ("DstMAC", 2)) (Modify ("OutPort", 2)) (Pred Drop))
    , [])
  ; ("combined 6"
    , mkIf (Match ("DstMAC", 1)) (Modify ("OutPort", 1))
     (mkIf (Match ("DstMAC", 2)) (Modify ("OutPort", 2)) (Pred Drop))
    , [])
  ; ("combined 7"
    , (Pred (Match ("DstMAC", 1)) <+> Pred (Match ("DstMAC", 2)))
      <.> Pred (Neg (Match ("DstMAC", 1)) <^> Neg (Match ("DstMAC", 2)))
    , [])
  ]
    
let make_with_tables
  (test_pols : (string * pol * ((TableName.t * pol) list)) list) =
  List.map
    (fun (name, p, blist) -> 
      let b = mkTableBinding blist in
      let p_tables, theta = replace_with_table p in
      let list_of_b = TableBinding.to_pol_list (theta b) in
      (name ^ " table", p_tables, list_of_b))
    test_pols

let test_pols = List.concat
  [ non_table_test_pols
  ; make_with_tables non_table_test_pols
  ]


(*** BENCHMARKS *******************************************************)

let max_ports = 16

let rec make_learning_switch_pols n : (pol * pol)=
  if n <= 0 then
    (Modify ("OutPort", -1), Pred Drop)
  else
    let port = n mod max_ports in
    let mpol, fwdpol = make_learning_switch_pols (n - 1) in
    let mpol' = mkIf ((Match ("SrcMAC", n))) (Pred Drop) mpol in
    let fwdpol' = mkIf ((Match ("DstMAC", n))) (Modify ("OutPort", port)) fwdpol in
    (mpol', fwdpol')

let learning_switch_without_tables size =
  let mpol, fpol = make_learning_switch_pols size in
  ("learning switch (no tables)", mpol <+> fpol, [])

let learning_switch size =
  let xmon = TableName.gen () in
  let xmon_t = (mkFieldSet ["SrcMAC"], mkFieldSet ["OutPort"]) in
  let xfwd = TableName.gen () in
  let xfwd_t = (mkFieldSet ["DstMAC"], mkFieldSet ["OutPort"]) in
  let pol = Table (xmon, xmon_t) <+> Table (xfwd, xfwd_t) in
  let lsmon, lsfwd = make_learning_switch_pols size in
  let b = [(xmon, lsmon); (xfwd, lsfwd)] in
  ("learning switch", pol, b)

(* Monitoring policy: remove unsafe MACs as they're seen.
 * Fwding policy: allow unsafe MACs as safe MACs connect to them.
 *)
let rec make_stateful_firewall_pols n : pol * pol =
  if n <= 0 then
    (Modify ("OutPort", -1), Pred Drop)
  else
    let mpol, fpol = make_stateful_firewall_pols (n - 1) in
    let mpol' = mkIf (Match ("DstMAC", n)) (Pred Drop) mpol in
    let fpol' = mkIf (Match ("SrcMAC", n)) (Pred Id) fpol in
    (mpol', fpol')

let stateful_firewall_without_tables size =
  let mpol, fpol = make_stateful_firewall_pols size in
  let pol =
    (Pred (Match ("InPort", 1)) <.> mpol)
    <+>
    (mkIf (Match ("InPort", 1))
      (Modify ("OutPort", 2))
      (fpol <.> Modify ("OutPort", 1))) in
  ("stateful firewall (no tables)", pol, [])

(* InPort 1 is the safe port.
 * InPort 2 is the unsafe port.
 *)
let stateful_firewall size =
  let xmon = TableName.gen () in
  let xmon_t = (mkFieldSet ["DstMAC"], mkFieldSet ["OutPort"]) in
  let xfwd = TableName.gen () in
  let xfwd_t = (mkFieldSet ["SrcMAC"], mkFieldSet ["OutPort"]) in
  let pol =
    (* Monitor new outgoing flows. *)
    (Pred (Match ("InPort", 1)) <.> Table (xmon, xmon_t))
    <+>
    (* Forward all outgoing flows.  Only forward incoming flows
     * that respond to previous outgoing flows. *)
    (mkIf (Match ("InPort", 1))
      (Modify ("OutPort", 2))
      (Table (xfwd, xfwd_t) <.> Modify ("OutPort", 1))) in
  let sfmon, sffwd = make_stateful_firewall_pols size in
  let b = [(xmon, sfmon); (xfwd, sffwd)] in
  ("stateful firewall", pol, b)

let benchmarks =
  [ learning_switch_without_tables 2
  ; learning_switch 2
  ; stateful_firewall_without_tables 2
  ; stateful_firewall 2
  ]


(*** LIST OF TESTS ********************************************************)

let mc_tests =
  [ ("multicast consolidation id", mctest_id)
  ; ("multicast consolidation drop", mctest_drop)
  ; ("multicast consolidation 1", mctest_1)
  ; ("multicast consolidation 2", mctest_2)
  ; ("multicast consolidation 3", mctest_3)
  ; ("multicast consolidation 4", mctest_4)
  ; ("multicast consolidation 5", mctest_5)
  ; ("multicast consolidation 6", mctest_6)
  ; ("multicast consolidation 7", mctest_7)
  ; mctest_8
  ; mctest_9
  ; mctest_10
  ; mctest_11
  ; mctest_12
  ; mctest_13
  ]

let fe_tests =
  [ fetest_id
  ; fetest_drop
  ; fetest_1
  ; fetest_2
  ; fetest_3
  ; fetest_3a
  ; fetest_4
  ; fetest_5
  ; fetest_6
  ; fetest_7
  ; fetest_8
  ; fetest_9
  ; fetest_10
  ; fetest_11
  ; fetest_12
  ; fetest_13
  ]

let tf_tests =
  [ ("blocks - 1", teblocks_1)
  ; ("table fitting single", tetest_single)
  ; ("tetest 1", tetest_1)
  ; ("tetest 2", tetest_2)
  ; ("tetest 3", tetest_3)
  ; ("tetest 4", tetest_4)
  ; ("tetest 5", tetest_5)
  ]

let combined_tests = mkTests test_pks test_pols 
let benchmark_tests = mkTests test_pks benchmarks 

let mc_tests = List.concat [mc_tests; fe_tests; tf_tests; combined_tests; benchmark_tests]

let tests = List.concat
  [ spec_tests
  ; mc_tests
  ]
