(*
 * The MIT License (MIT)
 * 
 * Copyright (c) 2015 Cole Schlesinger
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *)

open Syntax
open Semantics
open Compiler
open My_core
open Test_utils

(* Utility function. *)
let mkSemTest (p : pol) (in_pks : PacketSet.t) (expected_pks : PacketSet.t) =
  fun () ->
  eval p in_pks >>= fun out_pks ->
  if PacketSet.equal out_pks expected_pks then
    Ok None
  else
    let msg =
      Printf.sprintf
        "got: (%s),  expected: (%s)"
        (print_packets out_pks)
        (print_packets expected_pks) in
    Ok (Some msg)

(* Semantics tests. *)
let semtest_id =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let p = Pred Id in
  mkSemTest p pks pks

let semtest_drop =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let out_pks = mkPacketSet [] in
  let p = Pred Drop in
  mkSemTest p pks out_pks

let semtest_1 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let out_pks = mkPacketSet [] in
  let p = Pred (Match ("SrcMAC", 3)) in
  mkSemTest p pks out_pks

let semtest_2 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let out_pks =
    mkPacketSet
      [mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]] in
  let p = Pred (Match ("SrcMAC", 2)) in
  mkSemTest p pks out_pks

let semtest_3 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let out_pks = mkPacketSet [] in
  let p = Pred (And (Match ("SrcMAC", 1), Match ("SrcMAC", 2))) in
  mkSemTest p pks out_pks

let semtest_4 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let p = Pred (Or (Match ("SrcMAC", 1), Match ("SrcMAC", 2))) in
  mkSemTest p pks pks

let semtest_5 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let p = Pred (Neg (And (Match ("SrcMAC", 1), Match ("SrcMAC", 2)))) in
  mkSemTest p pks pks

let semtest_6 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let out_pks = mkPacketSet [] in
  let p = Pred (Neg (Or (Match ("SrcMAC", 1), Match ("SrcMAC", 2)))) in
  mkSemTest p pks out_pks


let semtest_10 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let expected_pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 3); ("DstMAC", 2) ]
      ] in
  let p = Seq (Pred (Match ("SrcMAC", 1)), (Modify ("SrcMAC", 3))) in
  mkSemTest p pks expected_pks

let semtest_11 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let expected_pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 3); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 3); ("DstMAC", 1) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ] in
  let p = Union (Pred (Match ("SrcMAC", 1)), (Modify ("SrcMAC", 3))) in
  mkSemTest p pks expected_pks

let semtest_12 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let expected_pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 3); ("DstMAC", 3) ]
      ] in
  let w1 = mkFieldSet [("SrcMAC")] in
  let w2 = mkFieldSet [("DstMAC")] in
  let p = Con ((Modify ("SrcMAC", 3)), w1, w2, (Modify ("DstMAC", 3))) in
  mkSemTest p pks expected_pks

let semtest_13 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let expected_pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ] in
  let p = IfThen (Match ("SrcMAC", 1), Pred Id, Pred Drop) in
  mkSemTest p pks expected_pks

let semtest_13a =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let expected_pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ] in
  let p = desugar_if ((Match ("SrcMAC", 1), Pred Id, Pred Drop)) in
  mkSemTest p pks expected_pks

let semtest_14 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let expected_pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let p = IfThen (Match ("SrcMAC", 1), Pred Drop, Pred Id) in
  mkSemTest p pks expected_pks

let semtest_15 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let expected_pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2); ("DstMAC", 1) ]
      ; mkPacket [ ("SrcMAC", 1); ("DstMAC", 1) ]
      ] in
  let p = Union (Modify ("SrcMAC", 1), Pred Id) in
  mkSemTest p pks expected_pks

let semtest_16 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ; mkPacket [ ("SrcMAC", 1) ]
      ] in
  let expected_pks = pks in
  let ws = FieldSet.empty in
  let p = Con (Pred Id, ws, ws, Pred Id) in
  mkSemTest p pks expected_pks

let semtest_17 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ; mkPacket [ ("SrcMAC", 1) ]
      ] in
  let expected_pks = PacketSet.empty in
  let p = (Pred (Match ("SrcMAC", 2))) <.> (Pred (Neg (Match ("SrcMAC", 2)))) in
  mkSemTest p pks expected_pks

let semtest_18 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ; mkPacket [ ("SrcMAC", 1) ]
      ] in
  let expected_pks = PacketSet.empty in
  let p = (Pred (Match ("SrcMAC", 2)))
    <.> (Pred (Neg (Match ("SrcMAC", 2)) <^> Neg (Match ("SrcMAC", 1)))) in
  mkSemTest p pks expected_pks

let semtest_19 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ; mkPacket [ ("SrcMAC", 1) ]
      ] in
  let expected_pks = PacketSet.empty in
  let p = (Pred (Match ("SrcMAC", 2)) <+> Pred (Match ("SrcMAC", 1)))
    <.> (Pred (Neg (Match ("SrcMAC", 2)) <^> Neg (Match ("SrcMAC", 1)))) in
  mkSemTest p pks expected_pks

let semtest_20 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ; mkPacket [ ("SrcMAC", 1) ]
      ] in
  let expected_pks = pks in
  let p = (Pred (Match ("SrcMAC", 2)) <+> Pred (Match ("SrcMAC", 1))) in
  mkSemTest p pks expected_pks

let semtest_21 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ; mkPacket [ ("SrcMAC", 1) ]
      ] in
  let expected_pks = mkPacketSet [mkPacket [ ("SrcMAC", 1) ]] in
  let p = (Pred (Match ("SrcMAC", 2)) <+> Pred (Match ("SrcMAC", 1)))
    <.> (Pred (Neg (Match ("SrcMAC", 2)))) in
  mkSemTest p pks expected_pks

let semtest_22 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ; mkPacket [ ("SrcMAC", 1) ]
      ] in
  let expected_pks = mkPacketSet [] in
  let p = Pred (Neg (Match ("SrcMAC", 2)) <^> Neg (Match ("SrcMAC", 1))) in
  mkSemTest p pks expected_pks

let semtest_23 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ; mkPacket [ ("SrcMAC", 1) ]
      ] in
  let expected_pks = mkPacketSet [] in
  let p = Pred (Neg (Match ("SrcMAC", 1)) <^> Neg (Match ("SrcMAC", 2))) in
  mkSemTest p pks expected_pks

let semtest_24 =
  let pks = mkPacketSet [] in
  let expected_pks = mkPacketSet [] in
  let p = Pred (Neg (Match ("SrcMAC", 1))) in
  mkSemTest p pks expected_pks

let semtest_25 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ] in
  let expected_pks = mkPacketSet [] in
  let p = Pred (Neg (Drop) <^> Neg (Id)) in
  mkSemTest p pks expected_pks

let semtest_26 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ] in
  let expected_pks = mkPacketSet [] in
  let p = Pred (Id <^> Drop) in
  mkSemTest p pks expected_pks

let semtest_27 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ] in
  let expected_pks = mkPacketSet [] in
  let p = Pred (Neg (Id)) in
  mkSemTest p pks expected_pks

let semtest_28 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ] in
  let expected_pks = mkPacketSet [] in
  let p = Pred (Id <^> Neg (Id)) in
  mkSemTest p pks expected_pks

let semtest_29 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ] in
  let expected_pks = mkPacketSet [] in
  let p = Pred (Neg Drop <^> Drop) in
  mkSemTest p pks expected_pks

let semtest_30 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ] in
  let expected_pks = pks in
  let p = Pred (Neg (Drop)) in
  mkSemTest p pks expected_pks

let semtest_31 =
  let pks =
    mkPacketSet
      [ mkPacket [ ("SrcMAC", 2) ]
      ] in
  let expected_pks = mkPacketSet [] in
  let p = Pred (Neg Id) in
  mkSemTest p pks expected_pks


let remove_fields_1 () : test_result =
  let pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 1) ]
      ] in
  let expected_pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("DstMAC", 1) ]
      ] in
  let w = mkFieldSet [("SrcMAC")] in
  let pks' = remove_fields w pks in
  if PacketSet.equal pks' expected_pks then
    Ok None
  else
    let msg =
      Printf.sprintf
        "got: (%s).  expected: (%s)."
        (print_packets pks')
        (print_packets expected_pks) in
    Ok (Some msg)


let cross_1 () : test_result =
  let pks1 =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2) ]
      ] in
  let pks2 =
    mkPacketSet
      [ mkPacket [ ("DstMAC", 2) ]
      ] in
  let expected_pks =
    mkPacketSet
      [ mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 1); ("DstMAC", 2) ]
      ; mkPacket [ ("EtherTyp", 0x806); ("SrcMAC", 2); ("DstMAC", 2) ]
      ] in
  cross pks1 pks2 >>= fun pks' ->
  if PacketSet.equal pks' expected_pks then
    Ok None
  else
    let msg =
      Printf.sprintf
        "got: (%s).  expected: (%s)."
        (print_packets pks')
        (print_packets expected_pks) in
    Ok (Some msg)

let tests =
  [ ("remove fields 1", remove_fields_1)
  ; ("cross fields 1", cross_1)
  ; ("sem test id", semtest_id)
  ; ("sem test drop", semtest_drop)
  ; ("sem test 1", semtest_1)
  ; ("sem test 2", semtest_2)
  ; ("sem test 3", semtest_3)
  ; ("sem test 4", semtest_4)
  ; ("sem test 5", semtest_5)
  ; ("sem test 6", semtest_6)
  ; ("sem test 10", semtest_10)
  ; ("sem test 11", semtest_11)
  ; ("sem test 12", semtest_12)
  ; ("sem test 13", semtest_13)
  ; ("sem test 13a", semtest_13a)
  ; ("sem test 14", semtest_14)
  ; ("sem test 15", semtest_15)
  ; ("sem test 16", semtest_16)
  ; ("sem test 17", semtest_17)
  ; ("sem test 18", semtest_18)
  ; ("sem test 19", semtest_19)
  ; ("sem test 20", semtest_20)
  ; ("sem test 21", semtest_21)
  ; ("sem test 22", semtest_22)
  ; ("sem test 23", semtest_23)
  ; ("sem test 24", semtest_24)
  ; ("sem test 25", semtest_25)
  ; ("sem test 26", semtest_26)
  ; ("sem test 27", semtest_27)
  ; ("sem test 28", semtest_28)
  ; ("sem test 29", semtest_29)
  ; ("sem test 30", semtest_30)
  ; ("sem test 31", semtest_31)
  ]
