(*
 * The MIT License (MIT)
 * 
 * Copyright (c) 2015 Cole Schlesinger
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *)

open Syntax
open Semantics
open My_core

(* Ok None: test passed!
 * Ok str : test failed, producing bad results.
 * Error  : test failed, because the underlying function failed.
 *)
type test_result = (string option, string) Result.t
type test_t = (unit -> test_result)

(* Infer a type t for p, then check that p type checks at t. *)
let infer_and_check (p : pol) : (pol_t, string) Result.t  =
  let p_t = infer_pol p in
  check_pol p p_t >>= fun (res : bool) ->
  match res with
  | true -> Ok p_t
  | false ->
    let msg =
      Printf.sprintf
        "policy does not type check at inferred type: (%s) : (%s)" 
        (print_pol p)
        (print_pol_t p_t) in
    Error msg

(* Build a packet (map) from field/int pairs. *)
let mkPacket (fields : (field * int) list) : Packet.t =
  List.fold_left
    (fun pk (f,v) -> FieldMap.add f v pk)
    Packet.empty
    fields

(* Used for zeroing metadata on incoming/outgoing packets. *)
let zero_metadata (pks : PacketSet.t) (fs : FieldSet.t) : PacketSet.t =
  let fs_map =
    mkPacket (List.map (fun f -> (f,0)) (FieldSet.elements fs)) in
  mkPacketSet (List.map
    (fun pk -> FieldMap.merge
      (fun _ a b -> 
        match a,b with
        | Some _, Some z -> Some z
        | Some a, _ -> Some a
        | _, Some b -> Some b
        | _, _ -> None)
      pk fs_map)
    (PacketSet.elements pks))

(* Evaluate both policies on pks and compare results after zeroing
 * out metadata. *)
let setupTestEval
  (pks : PacketSet.t)
  (metadata : FieldSet.t)
  (p : pol)
  (p_out : pol) : test_result =
  let zeroed_pks = zero_metadata pks metadata in
  eval p pks >>= fun pks_source ->
  eval p_out zeroed_pks >>= fun pks_target ->
  let zeroed_pks_source = zero_metadata pks_source metadata in
  let zeroed_pks_target = zero_metadata pks_target metadata in
  if PacketSet.equal zeroed_pks_source zeroed_pks_target then
    Ok None
  else
    let msg =
      Printf.sprintf "in: (%s), source out: (%s), target out: (%s)"
        (print_packets pks)
        (print_packets zeroed_pks_source)
        (print_packets zeroed_pks_target) in
    Ok (Some msg)
